# -*- coding: utf-8 -*-
"""
Functions relating velocity trend extrapolation
"""
from __future__ import (absolute_import, division, print_function,
                        unicode_literals)

__author__ = "yuhao"

import numpy as np

from pygeopressure.basic.well_log import Log
# from ..well_log import Log
v0 = 1600  # take values larger than 1500


def set_v0(v):
    """
    set global variable v0 for slotnick()
    """
    global v0
    v0 = v


def normal(x, a, b):
    r"""
    Extrapolate velocity using normal trend.

    Parameters
    ----------
    x : 1-d ndarray
        depth to convert
    a, b : scalar
        coefficents

    Returns
    -------
    out : 1-d ndarray
        esitmated velocity

    Notes
    -----
    .. math:: \log d{t}_{Normal}=a-bz

    is transformed to

    .. math:: v={e}^{bz-a}

    **Note** that the exponential relation is unphysical especially in depth
    bellow the interval within which the equation is calibrated.

    References
    ----------
    .. [1] C. Hottmann, R. Johnson, and others, "Estimation of formation
       pressures from log-derived shale properties," Journal of Petroleum
       Technology, vol. 17, no. 6, pp. 717-722, 1965.

    """
    return np.exp(x*b - a)


def normal_log(vel_log, a, b):
    """
    Returns
    -------
    Log
        normal velocity log
    """

    normal_vel = normal(np.array(vel_log.depth), a, b)
    mask = np.isnan(np.array(vel_log.data))
    normal_vel[mask] = np.nan
    log = Log()
    log.depth = np.array(vel_log.depth)
    log.data = normal_vel
    log.name = 'normal_vel_log'
    log.descr = "Velocity_normal"
    log.units = "m/s"

    return log


def slotnick(x, k):
    """
    Relation between velocity and depth

    Parameters
    ----------
    x : 1-d ndarray
        Depth to convert
    k : scalar
        velocity gradient

    Notes
    -----
    typical values of velocity gradient k falls in the range 0.6-1.0s-1

    References
    ----------
    .. [1] M. Slotnick, "On seismic computations, with applications, I,"
       Geophysics, vol. 1, no. 1, pp. 9-22, 1936.
    """
    global v0
    return v0 + k*x


def normal_dt(x, a, b):
    """
    normal trend of transit time

    Parameters
    ----------
    x : 1-d ndarray
        depth to convert
    """
    return a - b * x



# -*- coding: utf-8 -*-
"""
Routines for multivariate pressure prediction

Created on Sep 20 2018
"""
from __future__ import (absolute_import, division, print_function,
                        unicode_literals)

__author__ = "yuhao"

import numpy as np


def multivariate_virgin(sigma, phi, vsh, a_0, a_1, a_2, a_3, B):
    """
    Calculate velocity using multivariate virgin curve

    Parameters
    ----------
    sigma : 1-d ndarray
        effective pressure
    phi : 1-d ndarray
        effective porosity
    vsh : 1-d ndarray
        shale volume
    a_0, a_1, a_2, a_3 : float
        coefficients of equation
    B : float
        effective pressure exponential

    Returns
    -------
    out : 1-d ndarray
        velocity array

    Notes
    -----
    .. math:: V = a_0 + a_1\\phi + a_2{V}_{sh} + a_3 {\\sigma}^{B}

    [5]_

    .. [5] Sayers, C., Smit, T., van Eden, C., Wervelman, R., Bachmann, B.,
       Fitts, T., et al. (2003). Use of reflection tomography to predict
       pore pressure in overpressured reservoir sands. In submitted for
       presentation at the SEG 2003 annual meeting.
    """
    return a_0 - a_1 * phi - a_2 * vsh + a_3 * sigma**B


def invert_multivariate_virgin(vel, phi, vsh, a_0, a_1, a_2, a_3, B):
    """
    Calculate effective stress using multivariate virgin curve

    Parameters
    ----------
    vel : 1-d ndarray
        velocity array whose unit is m/s.
    phi : 1-d ndarray
        porosity array
    vsh : 1-d ndarray
        shale volume
    a_0, a_1, a_2, a_3 : scalar
        coefficients

    Returns
    -------
    sigma: 1-d ndarray
    """
    return ((vel - a_0 + a_1 * phi + a_2 * vsh) / a_3)**(1 / B)


def multivariate_unloading(sigma, phi, vsh, a_0, a_1, a_2, a_3, B, U, vmax):
    """
    Calculate velocity using multivariate unloading curve
    """
    sigma_max = invert_multivariate_virgin(
        vmax, phi, vsh, a_0, a_1, a_2, a_3, B)

    sigma_vc = sigma_max*(sigma/sigma_max)**(1/U)

    return multivariate_virgin(sigma_vc, phi, vsh, a_0, a_1, a_2, a_3, B)


def invert_multivariate_unloading(vel, phi, vsh, a_0, a_1, a_2, a_3,
                                  B, U, vmax):
    """
    Calculate effective stress using multivariate unloading curve
    """
    sigma_max = invert_multivariate_virgin(
        vmax, phi, vsh, a_0, a_1, a_2, a_3, B)

    sigma_vc = invert_multivariate_virgin(
        vel, phi, vsh, a_0, a_1, a_2, a_3, B)

    return sigma_max * (sigma_vc/sigma_max)**U


def effective_stress_multivariate(vel, phi, vsh, a_0, a_1, a_2, a_3,
                                  B, U, vmax, start_idx, end_idx=None):
    ves = invert_multivariate_virgin(vel, phi, vsh, a_0, a_1, a_2, a_3, B)
    unloading = invert_multivariate_unloading(
        vel, phi, vsh, a_0, a_1, a_2, a_3, B, U, vmax)
    ves[start_idx: end_idx] = unloading[start_idx: end_idx]
    return ves


def pressure_multivariate(obp, vel, phi, vsh, a_0, a_1, a_2, a_3,
                          B, U, vmax, start_idx, end_idx=None):
    """
    Pressure Prediction using multivariate model
    """
    ves = effective_stress_multivariate(
        vel, phi, vsh, a_0, a_1, a_2, a_3, B, U, vmax, start_idx, end_idx)
    return obp - ves


def pressure_multivariate_varu(obp, vel, phi, vsh, a_0, a_1, a_2, a_3,
                               B, U, vmax, start_idx, buf=20,
                               end_idx=None, end_buffer=10):
    """
    Pressure Prediction using multivariate model
    """
    ves = effective_stress_multivariate_varu(
        vel, phi, vsh, a_0, a_1, a_2, a_3,
        B, U, vmax, start_idx, buf, end_idx, end_buffer)
    return obp - ves


def effective_stress_multivariate_varu(vel, phi, vsh, a_0, a_1, a_2, a_3,
                                       B, U, vmax, start_idx, buf=20,
                                       end_idx=None, end_buffer=10):
    u_array = np.ones(vel.shape)
    u_array[start_idx: end_idx] = U
    # start buffer
    u_buffer = np.linspace(1, U, buf)
    u_array[start_idx-buf+1: start_idx + 1] = u_buffer
    # end buffer
    if end_idx is not None:
        u_array[end_idx: end_idx + end_buffer] = np.linspace(U, 1, end_buffer)
    ves = invert_multivariate_unloading(vel, phi, vsh, a_0, a_1, a_2, a_3,
                                        B, u_array, vmax)
    return ves



# -*- coding: utf-8 -*-
"""
some utilities
"""
from __future__ import (absolute_import, division, print_function,
                        unicode_literals)

from builtins import range

__author__ = "yuhao"

import sys

import numpy as np

PY_VER = sys.version_info.major

if PY_VER == 2:
    from functools import update_wrapper
    from singledispatch import singledispatch
else:
    from functools import singledispatch, update_wrapper


def rmse(measure, predict):
    """
    Relative Root-Mean-Square Error

    with RMS(y - y*) as nominator, and RMS(y) as denominator
    """
    delta = np.sqrt(np.mean((measure - predict)**2))
    denominator = np.sqrt(np.mean(measure**2))
    return delta/denominator

def nmse(measure, predict):
    """
    Normalized Root-Mean-Square Error

    with RMS(y - y*) as nominator, and MEAN(y) as denominator
    """
    delta = np.sqrt(np.mean((measure - predict)**2))
    denominator = np.mean(measure)
    return delta/denominator

def split_sequence(sequence, length):
    """
    Split a sequence into fragments with certain length
    """
    n_seq = len(sequence)
    for i in range(0, n_seq, length):
        yield sequence[i: i+length]


def methdispatch(func):
    dispatcher = singledispatch(func)
    def wrapper(*args, **kw):
        return dispatcher.dispatch(args[1].__class__)(*args, **kw)
    wrapper.register = dispatcher.register
    update_wrapper(wrapper, func)
    return wrapper


def pick_sparse(a_array, n):
    """
    Pick n equally spaced samples from array

    Parameters
    ----------
    a_array : 1-d ndarray
    n : int
        number of samples to pick
    """
    length = a_array.shape[0]
    if length < n:
        raise Exception("length of array smaller than n")
    step = length // (n - 1)
    if step == 0:
        step = 1
    new_list = []
    for i in range(0, n, step):
        new_list.append(a_array[i])
    new_array = a_array[::step]
    if length % (n - 1) != 0:
        return np.append(new_array, [a_array[-1]])
    else:
        return new_array



# -*- coding: utf-8 -*-
"""
class Log for well log data

Created on Fri Apr 18 2017
"""
from __future__ import (absolute_import, division, print_function,
                        unicode_literals)

from builtins import zip, open

__author__ = "yuhao"

import numpy as np
import matplotlib.pyplot as plt


class Log(object):
    """
    class for well log data
    """
    def __init__(self, file_name=None, log_name="unk"):
        """
        Parameters
        ----------
        file_name : str
            pseudo las file path
        log_name : str
            log name to create
        """
        self.name = log_name
        self.units = ""
        self.descr = ""
        self.prop_type = None
        self.__data = []
        self.__depth = []
        self.log_start = None
        self.log_stop = None
        self.depth_start = None
        self.depth_stop = None
        self.log_start_idx = None
        self.log_stop_idx = None
        if file_name is not None:
            self.__init_from_file(file_name)

    @classmethod
    def from_scratch(cls, depth, data, name=None, units=None, descr=None,
                     prop_type=None):
        log = cls()
        log.depth = np.array(depth)
        log.data = np.array(data)
        log.name = name
        log.units = units
        log.descr = descr
        log.prop_type = prop_type

        return log

    def __init_from_file(self, file_name):
        self._read_od(file_name)
        try:
            shorthand = self.descr[:3].lower()
            self.name = shorthand + "_unk"
            prop_dict = {
                'vel': 'VEL',
                'den': 'DEN',
                'sha': 'VSH',
                'ove': 'PRE',
                'pre': 'PRE'
            }
            try:
                self.prop_type = prop_dict[shorthand]
            except KeyError:
                pass
        except IndexError:
            self.name = "unk_unk"

    def __len__(self):
        return len(self.__data)

    def __str__(self):
        return "Well_Log:{}({}[{}])".format(self.name, self.descr, self.units)

    def __bool__(self):
        return bool(bool(self.__depth) and bool(self.__data))

    def __eq__(self, other):
        return self.depth == other.depth and self.data == other.data

    @property
    def depth(self):
        "depth data of the log"
        return list(self.__depth)

    @depth.setter
    def depth(self, values):
        self.__depth = list(values)

    @property
    def data(self):
        "property data of the log"
        return list(self.__data)

    @data.setter
    def data(self, values):
        self.__data = list(values)

    @property
    def start(self):
        "start depth of available property data"
        if self.log_start is None:
            for dep, dat in zip(self.__depth, self.__data):
                if np.isfinite(dat):
                    self.log_start = dep
                    break
        return self.log_start

    @property
    def start_idx(self):
        "start index of available property data"
        if self.log_start_idx is None:
            self.__data = np.array(self.__data)
            mask = np.isfinite(self.__data)
            index = np.where(mask == True)
            self.log_start_idx = index[0][0]
        return self.log_start_idx

    @property
    def stop(self):
        "end depth of available property data"
        if self.log_stop is None:
            for dep, dat in zip(reversed(self.__depth), reversed(self.__data)):
                if np.isfinite(dat):
                    self.log_stop = dep
                    break
        return self.log_stop

    @property
    def stop_idx(self):
        "end index of available property data"
        if self.log_stop_idx is None:
            self.__data = np.array(self.__data)
            mask = np.isfinite(self.__data)
            index = np.where(mask == True)
            self.log_stop_idx = index[0][-1] + 1
            # so when used in slice, +1 will not needed.
        return self.log_stop_idx

    @property
    def top(self):
        "top depth of this log"
        return self.__depth[0]

    @property
    def bottom(self):
        "bottom depth of this log"
        return self.__depth[-1]

    def _read_od(self, file_name):
        try:
            with open(file_name, "r") as fin:
                info_list = fin.readline().split('\t')
                temp_list = info_list[-1].split('(')
                self.descr = temp_list[0]
                self.units = temp_list[1][:-2]
                for line in fin:
                    tempList = line.split()
                    self.__depth.append(round(float(tempList[0]), 1))
                    if tempList[1] == "1e30":
                        self.__data.append(np.nan)
                    else:
                        self.__data.append(float(tempList[1]))
        except Exception as inst:
            print('{}: '.format(self.name))
            print(inst.args)

    def to_las(self, file_name):
        """
        Save as pseudo-las file
        """
        try:
            with open(file_name, 'w') as fout:
                split_list = self.descr.split(' ')
                description = '_'.join(split_list)
                fout.write("Depth(m)\t" + description + "(" + self.units + ")\n")
                for d, v in zip(self.__depth, self.__data):
                    d = str(d)
                    v = str(v) if np.isfinite(v) else "1e30"
                    fout.write("\t".join([d, v]) + "\n")
        except Exception as inst:
            print(inst.args)

    def get_depth_idx(self, d):
        "return index of depth"
        if d > self.bottom or d < self.top:
            return None
        else:
            return int((d - self.top) // 0.1)

    def get_data(self, depth):
        "get data at certain depth"
        depth_idx = list()
        for de in depth:
            depth_idx.append(self.get_depth_idx(de))
        log_depth = np.array(self.__depth)
        log_data = np.array(self.__data)
        mask = log_depth < 0
        for idx in depth_idx:
            if idx is not None:
                mask[idx] = True
        return log_data[mask]

    def get_resampled(self, rate):
        "return resampled log"
        standard_log_step = 0.1
        step = int(rate // standard_log_step) + 1
        log = Log()
        log.depth = self.depth[::step]
        log.data = self.data[::step]
        return log

    def plot(self, ax=None, color='gray', linewidth=0.5, linestyle='-',
             label=None, zorder=1):
        """
        Plot log curve

        Parameters
        ----------
        ax : matplotlib.axes._subplots.AxesSubplot
            axis object to plot on, a new axis will be created if not provided

        Returns
        -------
        matplotlib.axes._subplots.AxesSubplot
        """
        if ax is None:
            _, ax = plt.subplots()
            ax.invert_yaxis()
        if label is None:
            label = self.descr
        ax.plot(self.data, self.depth, linewidth=linewidth, color=color,
                linestyle=linestyle, label=label, zorder=zorder)
        ax.set(xlabel="{}({})".format(self.descr, self.units),
               ylabel="Depth(m)",
               title=self.name)
        return ax



# -*- coding: utf-8 -*-
"""
class for survey index definition

created on Jun 10th 2017
"""
from __future__ import (absolute_import, division, print_function,
                        unicode_literals)

__author__ = "yuhao"


class SurveyIndex(object):
    def __init__(self, value):
        self.value = int(value)

class InlineIndex(SurveyIndex):
    pass

class CrlineIndex(SurveyIndex):
    pass

class DepthIndex(SurveyIndex):
    def __init__(self, value):
        self.value = value

class CdpIndex(SurveyIndex):
    def __init__(self, cdp):
        try:
            self.inline, self.crline = [int(_) for _ in cdp]
            self.value = (self.inline, self.crline)
        except TypeError:
            raise TypeError("Expected tuple, got {}".format(type(cdp)))



# -*- coding: utf-8 -*-
"""
Routines for eaton seismic pressure prediction

Created on Sep 24 2018
"""
from __future__ import (absolute_import, division, print_function,
                        unicode_literals)

__author__ = "yuhao"

import numpy as np
from pygeopressure.basic.indexes import InlineIndex
from pygeopressure.basic.optimizer import optimize_nct_trace
from pygeopressure.velocity.extrapolate import normal
from pygeopressure.pressure.hydrostatic import hydrostatic_trace
from pygeopressure.pressure.utils import create_seis, create_seis_info
from pygeopressure.pressure.eaton import sigma_eaton


def eaton_seis(output_name, obp_cube, vel_cube, n,
               a=None, b=None, upper=None, lower=None):
    # create seismic object
    eaton_cube = create_seis(output_name, vel_cube)
    # create info file
    create_seis_info(eaton_cube, output_name)

    depth = np.array(list(vel_cube.depths()))

    hydrostatic = hydrostatic_trace(depth)
    hydro_inline = np.tile(hydrostatic, (vel_cube.nNorth, 1))

    # actual calcualtion
    for inl in vel_cube.inlines():
        obp_data_inline = obp_cube.data(InlineIndex(inl))
        vel_data_inline = vel_cube.data(InlineIndex(inl))
        vn_inline = np.zeros((vel_cube.nNorth, vel_cube.nDepth))
        for i, crl in enumerate(vel_cube.crlines()):
            cdp = (inl, crl)
            start_depth = upper.get_cdp(cdp)
            end_depth = lower.get_cdp(cdp)

            a, b = optimize_nct_trace(
                depth, vel_data_inline[i], start_depth, end_depth)

            vn_inline[i] = normal(depth, a, b)
        eaton_inline = obp_data_inline - \
            sigma_eaton(
                obp_data_inline-hydro_inline, vel_data_inline/vn_inline, n)

        eaton_cube.update(InlineIndex(inl), eaton_inline)

    return eaton_cube



# -*- coding: utf-8 -*-
"""
a Well class utilizing pandas DataFrame and hdf5 storage

Created on May 27 2018
"""
from __future__ import (absolute_import, division, print_function,
                        unicode_literals)

from builtins import zip, bytes, str

__author__ = "yuhao"

import json
from collections import OrderedDict
import random

import numpy as np
import pandas as pd
from scipy.optimize import curve_fit
from scipy.interpolate import interp1d

import matplotlib as mpl
from pygeopressure.pressure.eaton import eaton, sigma_eaton, power_eaton
from pygeopressure.pressure.bowers import (
    virgin_curve, invert_virgin, unloading_curve)
from pygeopressure.pressure.multivariate import multivariate_virgin
from pygeopressure.velocity.extrapolate import normal
from pygeopressure.basic.well_log import Log
from pygeopressure.basic.utils import rmse, pick_sparse


class LoadingPlot(object):
    """
    Parameters
    ----------
    json_file : str
        path to parameter file
    """
    def __init__(self, ax, obp_logs, vel_logs, pres_logs, well_names):
        self.ax = ax
        self.obp_logs = obp_logs
        self.vel_logs = vel_logs
        self.pres_logs = pres_logs
        self.well_names = well_names
        self.a = None
        self.b = None
        self._calculate_data()
        self._init_axis()

    def _init_axis(self):
        # self.ax.cla()
        # self.ax.set(title="{}".format("Well"),
        #             xlabel="Effective Stress(MPa)",
        #             ylabel="Velocity(m/s)")
        # self.ax.set_xlim(xmin=0, xmax=80)
        pass

    def _calculate_data(self):
        self.vels = []
        self.ess = []
        for obp_log, vel_log, pres_log in zip(self.obp_logs, self.vel_logs, self.pres_logs):
            vel = list()
            obp = list()
            pres = list()
            depth = np.array(vel_log.depth)
            for dp in pres_log.depth:
                idx = np.searchsorted(depth, dp)
                vel.append(vel_log.data[idx])
                obp.append(obp_log.data[idx])

            vel, obp, pres = np.array(vel), np.array(obp), np.array(pres_log.data)
            es = obp - pres
            self.vels.append(vel)
            self.ess.append(es)

    def plot(self):
        # self._init_axis()
        colors = list(mpl.colors.cnames.keys())
        random.seed(2018)
        for es, vel, name in zip(self.ess, self.vels, self.well_names):
            co = random.choice(colors)
            self.ax.scatter(es, vel, color=co, marker='d', label=name)
        self.ax.legend(loc=4)
        self.ax.figure.canvas.draw()

    def fit(self):
        popt, _ = curve_fit(
            virgin_curve, np.concatenate(self.ess), np.concatenate(self.vels))
        a, b = popt
        self.a, self.b = popt
        new_es = np.arange(0, 81)
        new_vel = virgin_curve(new_es, a, b)
        self.ax.plot(new_es, new_vel, color='gray', zorder=0)
        self.ax.figure.canvas.draw()

    def error_sigma(self):
        new_es = np.arange(0, 81)
        new_vel = virgin_curve(new_es, self.a, self.b)
        error_dict = {}
        f = interp1d(new_vel, new_es, kind='cubic')
        for vel, es, wn in zip(self.vels, self.ess, self.well_names):
            predict_es = f(vel)
            error_dict[wn] = (predict_es - es) / es * 100
        return error_dict

    def check_error(self, obp_log, vel_log, pres_log):
        # for obp_log, vel_log, pres_log in zip(self.obp_logs, self.vel_logs, self.pres_logs):
        vel = list()
        obp = list()
        pres = list()
        depth = np.array(vel_log.depth)
        for dp in pres_log.depth:
            idx = np.searchsorted(depth, dp)
            vel.append(vel_log.data[idx])
            obp.append(obp_log.data[idx])

        vel, obp, pres = np.array(vel), np.array(obp), np.array(pres_log.data)
        es = obp - pres

        new_es = np.arange(0, 81)
        new_vel = virgin_curve(new_es, self.a, self.b)
        # error_dict = {}
        f = interp1d(new_vel, new_es, kind='cubic')

        predict_es = f(vel)
        return (predict_es - es) / es * 100


def plot_bowers_vrigin(ax, a, b, well, vel_log, obp_log, upper, lower,
                       pres_log='loading', mode='nc', nnc=5):
    if isinstance(upper, (bytes, str)):
        depth_upper = well.params['horizon'][upper]
    else:
        depth_upper = upper
    if isinstance(upper, (bytes, str)):
        depth_lower = well.params['horizon'][lower]
    else:
        depth_lower = lower
    if isinstance(vel_log, (bytes, str)):
        vel_log = well.get_log(vel_log)
    if isinstance(obp_log, (bytes, str)):
        obp_log = well.get_log(obp_log)
    if isinstance(pres_log, (bytes, str)):
        # pres_log = well.get_loading_pressure()
        pres_log = well.get_pressure(pres_log)

    depth = np.array(obp_log.depth)

    nct_vel_to_fit = []
    nct_es_to_fit = []
    pres_vel_to_fit = []
    pres_es_to_fit = []
    if mode == 'nct' or mode == 'both':
        nct_es_data = np.array(obp_log.data) - np.array(well.hydrostatic)
        nct_mask = depth < depth_lower
        nct_mask *= depth > depth_upper
        nct_mask *= depth < vel_log.stop

        nct_vel_interval = np.array(vel_log.data)[nct_mask]
        nct_es_interval = nct_es_data[nct_mask]

        nct_vel_to_fit = np.array(pick_sparse(nct_vel_interval, nnc))
        nct_es_to_fit = np.array(pick_sparse(nct_es_interval, nnc))
        ax.scatter(
            nct_es_to_fit, nct_vel_to_fit, color='blue', marker='d',
            label='NCP')
    if mode == 'pres' or mode == 'both':
        vel = list()
        obp = list()
        pres = list()
        for dp in pres_log.depth:
            idx = np.searchsorted(depth, dp)
            vel.append(vel_log.data[idx])
            obp.append(obp_log.data[idx])
        vel, obp, pres = np.array(vel), np.array(obp), np.array(pres_log.data)
        es = obp - pres

        pres_vel_to_fit = vel
        pres_es_to_fit = es
        ax.scatter(
            pres_es_to_fit, pres_vel_to_fit, color='purple', marker='s',
            label='measured')

    es_curve = np.arange(0, 80, 1)
    vel_curve = virgin_curve(es_curve, a, b)
    ax.plot(es_curve, vel_curve, color='black', zorder=1, label="Loading")

    ax.set(title="Loading Curve - {}".format(well.well_name),
           xlabel="Effective Stress(MPa)",
           ylabel="Velocity(m/s)")
    ax.set_xlim(left=0)
    ax.set_ylim(bottom=1500)

    ax.legend(loc=4)


def plot_bowers_unloading(ax, a, b, u, vmax, well, vel_log, obp_log,
                          pres_log='unloading'):
    """
    plot bowers unloading plot
    """
    if isinstance(vel_log, (bytes, str)):
        vel_log = well.get_log(vel_log)
    if isinstance(obp_log, (bytes, str)):
        obp_log = well.get_log(obp_log)
    if isinstance(pres_log, (bytes, str)):
        pres_log = well.get_pressure(pres_log)

    depth = np.array(obp_log.depth)

    vel = list()
    obp = list()
    pres = list()
    for dp in pres_log.depth:
        idx = np.searchsorted(depth, dp)
        vel.append(vel_log.data[idx])
        obp.append(obp_log.data[idx])
    vel, obp, pres = np.array(vel), np.array(obp), np.array(pres_log.data)
    es = obp - pres

    ax.scatter(es, vel, marker="^", color='r', label='meassured')

    sigma_max = invert_virgin(vmax, a, b)

    es_curve = np.arange(0, sigma_max+1, 1)


    vel_curve = unloading_curve(es_curve, a, b, u, vmax)

    ax.plot(es_curve, vel_curve, label='unloading')
    ax.legend(loc='lower right')


def plot_eaton_error(ax, well, vel_log, obp_log, a, b, pres_log="loading"):
    if isinstance(vel_log, (bytes, str)):
        vel_log = well.get_log(vel_log)
    if isinstance(obp_log, (bytes, str)):
        obp_log = well.get_log(obp_log)
    if isinstance(pres_log, (bytes, str)):
        # pres_log = well.get_loading_pressure()
        pres_log = well.get_pressure(pres_log)

    depth = np.array(obp_log.depth)

    hydrostatic = np.array(well.hydrostatic)
    es_normal = np.array(obp_log.data) - hydrostatic
    v_normal = normal(depth, a, b)

    vel = list()
    vel_norm = list()
    es_norm = list()
    obp = list()
    pres = list()
    for dp in pres_log.depth:
        idx = np.searchsorted(depth, dp)
        vel.append(vel_log.data[idx])
        vel_norm.append(v_normal[idx])
        es_norm.append(es_normal[idx])
        obp.append(obp_log.data[idx])
    vel, vel_norm = np.array(vel), np.array(vel_norm)
    vel_ratio = vel / vel_norm

    obp, pres = np.array(obp), np.array(pres_log.data)
    es = obp - pres
    es_norm = np.array(es_norm)
    es_ratio = es / es_norm

    popt, _ = curve_fit(power_eaton, vel_ratio, es_ratio)
    n, = popt

    rms_err = rmse(es, sigma_eaton(es_norm, vel_ratio, n))

    n_array = np.arange(n-2, n+2, 0.01)
    err_array = []
    for ni in n_array:
        predict_sigma = sigma_eaton(es_norm, vel_ratio, ni)
        err_array.append(rmse(es, predict_sigma))

    ax.plot(n_array, err_array, color='black')
    ax.axvline(x=n, color='g')
    ax.set(title='Optimize Eaton', xlabel="n", ylabel="RMS Error")
    info_string = "n:{}\nRMS:{}".format(n, rms_err)
    ax.text(
        s="{}".format(info_string), x=0.1, y=0.1, color='b',
        transform=ax.transAxes, size=9)


def plot_multivariate(axes, well, vel_log, por_log, vsh_log, obp_log,
                      upper, lower, a0, a1, a2, a3, B):

    axes[0].plot(np.array(vel_log.data)/1000, vel_log.depth, linewidth=0.5,
                 color='gray')

    axes[0].set(xlabel='Vp (km/s)', ylabel='Depth (m)', ylim=[lower, upper])


    axes[1].plot(por_log.data, por_log.depth, linewidth=0.5, color='gray')
    axes[1].set(xlabel='$\phi$')

    axes[2].plot(vsh_log.data, vsh_log.depth, linewidth=0.5, color='gray')
    axes[2].set(xlabel='$V_{sh}$')

    depth = well.depth
    hydrostatic = well.hydrostatic
    obp = well.lithostatic
    es = obp - hydrostatic

    axes[3].plot(es**B, depth, color='gray')
    axes[3].set(xlabel='${\sigma}^{B}$')

    vel_predict = multivariate_virgin(
        es, np.array(por_log.data), np.array(vsh_log.data), a0, a1, a2, a3, B)
    axes[0].plot(vel_predict/1000, vel_log.depth)



# -*- coding: utf-8 -*-
"""
Routines for Bowers' pore pressure prediction with seismic velocity
"""
from __future__ import (absolute_import, division, print_function,
                        unicode_literals)

__author__ = "yuhao"

import numpy as np

from pygeopressure.basic.indexes import InlineIndex, CdpIndex
from pygeopressure.basic.optimizer import optimize_bowers_trace
from pygeopressure.pressure.bowers import invert_virgin
from pygeopressure.pressure.hydrostatic import hydrostatic_trace
from pygeopressure.pressure.utils import create_seis, create_seis_info


def bowers_seis(output_name, obp_cube, vel_cube, a=None, b=None,
                upper=None, lower=None, mode='simple'):
    # create seismic object
    bowers_cube = create_seis(output_name, vel_cube)
    # create info file
    create_seis_info(bowers_cube, output_name)
    # calculation
    if mode == 'optimize':
        # with optimization
        bowers_optimize(bowers_cube, obp_cube, vel_cube, upper, lower)
    else:
        # simple
        bowers_simple(bowers_cube, obp_cube, vel_cube, a, b)

    return bowers_cube


def bowers_simple(bowers_cube, obp_cube, vel_cube, a=None, b=None):
    """
    Bowers prediction with fixed a, b
    """
    for inl in vel_cube.inlines():
        obp_data_inline = obp_cube.data(InlineIndex(inl))
        vel_data_inline = vel_cube.data(InlineIndex(inl))

        bowers_inline = obp_data_inline - \
            invert_virgin(vel_data_inline, a, b)

        bowers_cube.update(InlineIndex(inl), bowers_inline)


def bowers_optimize(bowers_cube, obp_cube, vel_cube, upper_hor, lower_hor):
    """
    Bowers prediction with automatic coefficient optimization
    """
    depth_tr = np.array(list(vel_cube.depths()))
    hydro_tr = hydrostatic_trace(depth_tr)

    for inl in vel_cube.inlines():
        bowers_data_inline = np.zeros((vel_cube.nNorth, vel_cube.nDepth))
        for i, crl in enumerate(vel_cube.crlines()):
            vel_tr = vel_cube.data(CdpIndex((inl, crl)))
            obp_tr = obp_cube.data(CdpIndex((inl, crl)))
            depth_upper = upper_hor.get_cdp((inl, crl))
            if lower_hor == "bottom":
                depth_lower = depth_tr[-1]
            else:
                depth_lower = lower_hor.get_cdp((inl, crl))
            try:
                a, b = optimize_bowers_trace(
                    depth_tr, vel_tr, obp_tr, hydro_tr,
                    depth_upper, depth_lower)
            except:
                raise Exception("cdp{},{}".format(inl, crl))
            bowers_data_inline[i] = invert_virgin(vel_tr, a, b)

        bowers_cube.update(InlineIndex(inl), bowers_data_inline)



# -*- coding: utf-8 -*-
"""
Created on Thu Apr 26 2017
"""
from __future__ import (absolute_import, division, print_function,
                        unicode_literals)

from builtins import range

__author__ = "yuhao"

import numpy as np
import matplotlib.pyplot as plt
from scipy.signal import cspline1d, cspline1d_eval
from matplotlib.colors import LinearSegmentedColormap, Normalize


class Wiggles(object):
    def __init__(self, data, wiggleInterval=10, overlap=1, posFill='black',
                 negFill=None, lineColor='black', rescale=True, extent=None, ax=None):
        self.data = data
        self.origin = 0
        self.posFill = 'black'
        self.negFill = None
        self.lineColor = None
        self.resampleRatio = 10
        self.rescale = True
        self.zmin = 0
        self.zmax = None
        self.ax = ax
        self.extent = extent
        self.wiggleInterval = wiggleInterval
        self.overlap = overlap

    def wiggle(self, values):
        """
        Plot a trace in VAWT(Variable Area Wiggle Trace)
        """
        if self.zmax is None:
            self.zmax = values.size

        # Rescale so that values ranges from -1 to 1
        if self.rescale:
            values = values.astype(np.float)
            values -= values.min()
            values /= values.ptp()
            values *= 2
            values -= 1

        # Interpolate at resampleRatio x the previous density
        resample_z = np.linspace(0, values.size, values.size * self.resampleRatio)
        # cubic spline interpolation
        cj = cspline1d(values)
        resample_v = cspline1d_eval(cj, resample_z)
        print(resample_v)
        newz = resample_z
        if self.origin is None:
            self.origin = resample_v.mean()

        # Plot
        if self.posFill is not None:
            self.ax.fill_betweenx(newz, resample_v, self.origin,
                                  where=resample_v > self.origin,
                                  facecolor=self.posFill)
        if self.negFill is not None:
            self.ax.fill_betweenx(newz, resample_v, self.origin,
                                  where=resample_v < self.origin,
                                  facecolor=self.negFill)
        if self.lineColor is not None:
            self.ax.plot(resample_v, newz, color=self.lineColor, linewidth=.1)

    def wiggles(self):
        """
        2-D Wiggle Trace Variable Amplitude Plot
        """
        # Rescale so that the data ranges from -1 to 1
        if self.rescale:
            self.data = self.data.astype(np.float)
            self.data -= self.data.min()
            self.data /= self.data.ptp()
            self.data *= 2
            self.data -= 1

        if self.extent is None:
            xmin, ymin = 0, self.data.shape[0]
            ymax, xmax = 0, self.data.shape[1]
        else:
            xmin, xmax, ymin, ymax = self.extent

        if self.ax is None:
            _, self.ax = plt.subplots()
        self.ax.invert_yaxis()
        self.ax.set(xlim=[xmin, xmax], ylim=[ymin, ymax]) # xrange should be larger!!!
        ny, nx = self.data.shape
        x_loc = np.linspace(xmin, xmax, nx)
        for i in range(self.wiggleInterval//2, nx, self.wiggleInterval):
            x = self.overlap * (self.wiggleInterval / 2.0) * (x_loc[1] - x_loc[0]) * self.data[:, i]
            self.wiggle(x + x_loc[i])


def wiggle(values, origin=0, posFill='black', negFill=None, lineColor='black',
           resampleRatio=10, rescale=False, zmin=0, zmax=None, ax=None):
    """
    Plot a trace in VAWT(Variable Area Wiggle Trace)

    Parameters
    ----------
    x: input data (1D numpy array)

    origin: (default, 0) value to fill above or below (float)

    posFill: (default, black)
        color to fill positive wiggles with (string or None)

    negFill: (default, None)
        color to fill negative wiggles with (string or None)

    lineColor: (default, black)
        color of wiggle trace (string or None)

    resampleRatio: (default, 10)
        factor to resample traces by before plotting (1 = raw data) (float)

    rescale: (default, False)
        If True, rescale "x" to be between -1 and 1

    zmin: (default, 0)
        The minimum z to use for plotting

    zmax: (default, len(x))
        The maximum z to use for plotting

    ax: (default, current axis)
        The matplotlib axis to plot onto

    Returns
    -------
    Plot
    """
    if zmax is None:
        zmax = values.size

    # Rescale so that values ranges from -1 to 1
    if rescale:
        values = values.astype(np.float)
        values -= values.min()
        values /= values.ptp()
        values *= 2
        values -= 1

    # Interpolate at resampleRatio x the previous density
    resample_z = np.linspace(0, values.size, values.size * resampleRatio)
    # cubic spline interpolation
    cj = cspline1d(values)
    resample_v = cspline1d_eval(cj, resample_z)

    # newz = np.linspace(zmax, zmin, resample_z.size)
    # newz = np.linspace(zmin, zmax, resample_z.size)
    newz = resample_z
    if origin is None:
        origin = resample_v.mean()

    # # Plot
    # if ax is None:
    #     ax = plt.gca()
    #     # plt.hold(True)
    if posFill is not None:
        ax.fill_betweenx(newz, resample_v, origin,
                         where=resample_v > origin,
                         facecolor=posFill)
    if negFill is not None:
        ax.fill_betweenx(newz, resample_v, origin,
                         where=resample_v < origin,
                         facecolor=negFill)
    if lineColor is not None:
        ax.plot(resample_v, newz, color=lineColor, linewidth=.1)


def wiggles(data, wiggleInterval=10, overlap=5, posFill='black',
            negFill=None, lineColor='black', rescale=True, extent=None, ax=None):
    """
    2-D Wiggle Trace Variable Amplitude Plot

    Parameters
    ----------
    x: input data (2D numpy array)
    wiggleInterval: (default, 10) Plot 'wiggles' every wiggleInterval traces
    overlap: (default, 0.7) amount to overlap 'wiggles' by (1.0 = scaled
            to wiggleInterval)
    posFill: (default, black) color to fill positive wiggles with (string
            or None)
    negFill: (default, None) color to fill negative wiggles with (string
            or None)
    lineColor: (default, black) color of wiggle trace (string or None)
    resampleRatio: (default, 10) factor to resample traces by before
            plotting (1 = raw data) (float)
    extent: (default, (0, nx, 0, ny)) The extent to use for the plot.
            A 4-tuple of (xmin, xmax, ymin, ymax)
    ax: (default, current axis) The matplotlib axis to plot onto.
    Output:
        a matplotlib plot on the current axes
    """
    # Rescale so that the data ranges from -1 to 1
    if rescale:
        data = data.astype(np.float)
        data -= data.min()
        data /= data.ptp()
        data *= 2
        data -= 1

    if extent is None:
        xmin, ymin = 0, data.shape[0]
        ymax, xmax = 0, data.shape[1]
    else:
        xmin, xmax, ymin, ymax = extent

    if ax is None:
        _, ax = plt.subplots()
    ax.invert_yaxis()
    ax.set(xlim=[xmin, xmax], ylim=[ymin, ymax]) # xrange should be larger!!!
    ny, nx = data.shape
    x_loc = np.linspace(xmin, xmax, nx)
    for i in range(wiggleInterval//2, nx, wiggleInterval):
        x = overlap * (wiggleInterval / 2.0) * (x_loc[1] - x_loc[0]) * data[:, i]
        wiggle(x + x_loc[i], origin=x_loc[i], posFill=posFill, negFill=negFill,
               lineColor=lineColor, zmin=ymin, zmax=ymax, ax=ax)
    # decorations
    ax.xaxis.set_ticks_position('both')
    ax.yaxis.set_ticks_position('both')
    ax.xaxis.set_tick_params(labeltop='on', labelbottom='off')
    # # ticks_list = [(0.2, 0.2), (0.4, 0.4), (0.6, 0.6), (0.8, 0.8)]
    # ticks_list = [(0.2,0.2),(1,1)]
    # ax.set(
    #     yticks=[item[1] for item in ax.transAxes.transform(ticks_list)],
    #     xticks=[item[0] for item in ax.transAxes.transform(ticks_list)]
    # )


def img(data, extent, ax, cm='seismic', ptype='seis'):
    data = np.nan_to_num(data)
    if cm == 'seismic_od':
        cm = opendtect_seismic_colormap()
    if cm in ('seismic', 'seismic_od'):
        if ptype == 'seis':
            mean_value = np.mean(data)
            std_value = np.std(data)
            vmin = mean_value - 2*std_value
            vmax = mean_value + 2*std_value
        else:
            vmin = np.min(data)
            vmax = np.max(data)
    else:
        vmin, vmax = None, None
    _ = ax.imshow(
        data,
        interpolation='bicubic',
        origin='lower',
        extent=extent,
        cmap=cm,
        aspect='auto',
        vmin=vmin,
        vmax=vmax)
    ax.tick_params(direction='out')
    ax.xaxis.set_ticks_position('both')
    ax.yaxis.set_ticks_position('both')
    ax.xaxis.set_tick_params(labeltop='on', labelbottom='off')

    return _


def opendtect_seismic_colormap():
    normalizer = Normalize(vmin=0, vmax=255)

    cdict = {
        'red': [
            (0, float(normalizer(170)), float(normalizer(170))),
            (0.070352, float(normalizer(255)), float(normalizer(255))),
            (0.25, float(normalizer(255)), float(normalizer(255))),
            (0.5, float(normalizer(243)), float(normalizer(243))),
            (0.88324, float(normalizer(56)), float(normalizer(56))),
            (1.0, 0.0, 0.0)
        ],
        'green': [
            (0.0, 0.0, 0.0),
            (0.070352, float(normalizer(28)), float(normalizer(28))),
            (0.25, float(normalizer(200)), float(normalizer(200))),
            (0.5, float(normalizer(243)), float(normalizer(243))),
            (0.88324, float(normalizer(70)), float(normalizer(70))),
            (1.0, 0.0, 0.0)
        ],
        'blue': [
            (0.0, 0.0, 0.0),
            (0.070352, 0, 0),
            (0.25, 0, 0),
            (0.5, float(normalizer(243)), float(normalizer(243))),
            (0.88324, float(normalizer(127)), float(normalizer(127))),
            (1.0, 0.0, 0.0)
        ]
    }

    return LinearSegmentedColormap('Seismic_OD', cdict)



from pygeopressure.basic.well import Well
from pygeopressure.basic.well_storage import WellStorage
from pygeopressure.basic.well_log import Log
from pygeopressure.basic.log_tools import (
    smooth_log, truncate_log, interpolate_log, upscale_log, local_average,
    shale, extrapolate_log_traugott)
from pygeopressure.basic.seisegy import SeiSEGY
from pygeopressure.basic.survey import Survey
from pygeopressure.basic.las import LasData
from pygeopressure.basic.survey_setting import SurveySetting
from pygeopressure.basic.threepoints import ThreePoints
from pygeopressure.basic.utils import rmse, nmse
from pygeopressure.basic.indexes import (
    InlineIndex, CrlineIndex, DepthIndex, CdpIndex)
from pygeopressure.basic.horizon import Horizon
from pygeopressure.basic.optimizer import (
    optimize_nct, optimize_eaton, optimize_bowers_virgin, optimize_traugott,
    optimize_bowers_unloading, optimize_multivaraite)
from pygeopressure.basic.plots import (
    plot_eaton_error, plot_bowers_vrigin, plot_bowers_unloading,
    plot_multivariate)
from pygeopressure.basic.survey import create_survey_directory

from pygeopressure.velocity.smoothing import smooth, smooth_2d, smooth_trace
from pygeopressure.velocity.conversion import (
    rms2int, int2rms, int2avg, avg2int, twt2depth)
from pygeopressure.velocity.interpolation import interp_DW, spline_1d
from pygeopressure.velocity.extrapolate import (
    set_v0, normal, slotnick, normal_dt)

from pygeopressure.pressure.obp import (
    overburden_pressure, obp_well, obp_trace, obp_seis, gardner, gardner_seis,
    traugott, traugott_trend)
from pygeopressure.pressure.bowers import (
    virgin_curve, invert_virgin, unloading_curve, invert_unloading,
    bowers, bowers_varu)
from pygeopressure.pressure.bowers_seis import bowers_seis
from pygeopressure.pressure.eaton import eaton
from pygeopressure.pressure.eaton_seis import eaton_seis
from pygeopressure.pressure.multivariate import (
    multivariate_virgin, invert_multivariate_virgin,
    multivariate_unloading, invert_multivariate_unloading,
    effective_stress_multivariate, pressure_multivariate,
    pressure_multivariate_varu)
from pygeopressure.pressure.hydrostatic import (
    hydrostatic_pressure, hydrostatic_well, hydrostatic_trace)

from ._version import get_versions
__version__ = get_versions()['version']
del get_versions



# -*- coding: utf-8 -*-
"""
Routines performing velocity type conversion
"""
from __future__ import (absolute_import, division, print_function,
                        unicode_literals)

from builtins import range

__author__ = "yuhao"

import numpy as np
from scipy import interpolate


def rms2int(twt, v_rms):
    r"""
    Convert rms velocity to interval velocity

    Parameters
    ----------
    twt : 1-d ndarray
        input two-way-time array, in ms
    rms : 1-d nadarray
        rms velocity array, in m/s

    Returns
    -------
    v_int : 1-d ndarray
        interval velocity array with the same length of twt and rms

    Notes
    -----
    This routine uses Dix equation to comput inverval velocity.

    .. math:: V_{int}[i]^2 = \frac{V_{rms}[i]^2 t_{i} - V_{rms}[i-1]^2 \
                t_{i-1}}{t_{i}-t_{i-1}}

    twt and rms should be of the same length of more than 2.

    Examples
    --------
    >>> a = np.arange(10)
    >>> twt = np.arange(10)
    >>> rms2int(twt, a)
    array([  0.        ,   1.        ,   2.64575131,   4.35889894,
             6.08276253,   7.81024968,   9.53939201,  11.26942767,
            13.        ,  14.73091986])
    """
    v_int = np.ones((len(twt), ))
    # v_int = np.copy(rms)
    twt = twt*0.001
    v_int[0] = v_rms[0]
    v_int[1:] = np.sqrt(
        (v_rms[1:]**2 * twt[1:] - v_rms[:-1]**2 * twt[:-1]) / \
        (twt[1:] - twt[:-1])
    )
    # for i in range(1, v_rms.shape[0]):
    #     v_int[i] = np.sqrt(
    #         (v_rms[i]**2 * twt[i] - v_rms[i-1]**2 * twt[i-1]) / \
    #         (twt[i] - twt[i-1])
    #     )

    return v_int


def int2rms(twt, v_int):
    """
    Parameters
    ----------
    twt : 1-d ndarray
    v_int : 1-d ndarray

    Returns
    -------
    v_rms : 1-d ndarray
    """
    v_rms = np.ones((len(twt), ))
    twt = twt*0.001
    v_rms[0] = v_int[0]
    # because of the accumulational effect of rms velocity, this func should
    # not use vectorized notation
    for i in range(1, v_int.shape[0]):
        v_rms[i] = np.sqrt(
            (v_int[i]**2 * (twt[i] - twt[i-1]) + v_rms[i-1]**2 * twt[i-1]) / twt[i]
        )

    return v_rms


def int2avg(twt, v_int):
    r"""
    Parameters
    ----------
    twt : 1-d ndarray

    v_int : 1-d ndarray

    Returns
    -------
    v_avg : 1-d ndarray

    Notes
    -----
    .. math:: V_{int}[i](t_{i} - t_{i-1}) = V_{avg}[i] t_{i} - \
              V_{avg}[i-1] t_{i-1}
    """
    v_avg = np.ones((len(twt), ))
    twt = twt * 0.001
    v_avg[0] = v_int[0]
    for i in range(1, v_int.shape[0]):
        v_avg[i] = (v_avg[i-1] * twt[i-1] + v_int[i] * (twt[i] - twt[i-1])) / \
            twt[i]
    return v_avg


def avg2int(twt, v_avg):
    """
    Parameters
    ----------
    twt : 1-d ndarray
    v_avg : 1-d ndarray

    Returns
    -------
    v_int : 1-d ndarray
    """
    v_int = np.ones((len(twt), ))
    twt = twt * 0.001
    v_int[0] = v_avg[0]
    v_int[1:] = (v_avg[1:]*twt[1:] - v_avg[:-1]*twt[:-1]) /\
        (twt[1:] - twt[:-1])

    return v_int


def twt2depth(twt, v_avg, prop_2_convert,
              stepDepth=4, startDepth=None, endDepth=None):
    """
    Parameters
    ----------
    twt : 1-d ndarray
    v_avg : 1-d ndarray
    prop_2_convert: 1-d ndarray
    stepDepth : scalar
    startDpeth (optional): scalar
    endDepth (optional): scalar

    Returns
    -------
    newDepth : 1-d ndarray
        new depth array
    new_prop_2_convert : 1-d ndarray
        average velocity in depth domain
    """
    depth = np.ones((len(twt), ))
    twt = twt * 0.001
    depth = twt * v_avg / 2.0
    startDepth = depth[0] if startDepth is None else startDepth
    endDepth = depth[-1] if endDepth is None else endDepth
    newDepth = np.arange(startDepth, endDepth+0.01, stepDepth)
    f = interpolate.interp1d(depth, prop_2_convert, kind='cubic')
    new_prop_2_convert = f(newDepth)

    return (newDepth, new_prop_2_convert)



# -*- coding: utf-8 -*-
"""
Function to calculate hydrostatic pressure

Created on Fri Nov 11 2016
"""
from __future__ import (absolute_import, division, print_function,
                        unicode_literals)

__author__ = "yuhao"

import numpy as np

from pygeopressure.basic.well_log import Log


def hydrostatic_pressure(depth, kelly_bushing=0, depth_w=0, rho_f=1., rho_w=1.):
    """
    Parameters
    ----------
    depth : scalar or 1-d ndarray
        measured depth, unit: meter
    rho_f : scalar
        density of pore fluid, g/cm3
    kelly_bushing : scalar
        kelly bushing elevation, in meter
    depth_w : scalar
        sea water depth
    rho_w : scalar
        sea water density

    Returns
    -------
    pressure : scalar or 1-d ndarray
        unit: mPa
    """
    depth = np.array(depth)
    rho_f *= 1000  # kg/m3
    rho_w *= 1000  # kg/m3
    acceleration = 9.80665  # m/s2
    density = np.full_like(depth, rho_f)
    # sea level to sea bottom
    mask = depth < (kelly_bushing + depth_w)
    density[mask] = rho_w
    # kelly bushing to sea level
    mask = depth < kelly_bushing
    density[mask] = 0
    delta_depth = np.full_like(depth, np.nan)
    delta_depth[1:] = depth[1:] - depth[:-1]
    delta_depth[0] = 0
    hydrostatic = delta_depth * density * acceleration / 1000000  # mPa
    hydrostatic = np.cumsum(hydrostatic)

    return hydrostatic


def hydrostatic_well(depth, kb=0, wd=0, rho_f=1., rho_w=1.):
    """
    Returns
    -------
    Log
        Hydrostatic pressure as a Log
    """
    hydrostatic = hydrostatic_pressure(
        depth, kelly_bushing=kb, depth_w=wd, rho_f=rho_f, rho_w=rho_w)

    hydro_log = Log()
    hydro_log.depth = np.array(depth)
    hydro_log.data = hydrostatic
    hydro_log.name = 'log_hydro'
    hydro_log.descr = "Hydrostatic_Pressure"
    hydro_log.units = "MPa"

    return hydro_log


def hydrostatic_trace(depth, rho=1.01, g=9.8, shift=0):
    mask = depth <= shift
    hydro = (depth - shift) * rho * g * 0.001
    hydro[mask] = 0
    return hydro



# -*- coding: utf-8 -*-
"""
2-d interpolation routines
"""
from __future__ import (absolute_import, division, print_function,
                        unicode_literals)

from builtins import range

__author__ = "yuhao"

from scipy import interpolate
import numpy as np


def interp_DW(array2d):
    """2-D distance-weighted interpolation

    Parameters
    ----------
    array2d : ndarray
        2-D ndarray void values being singaled by np.nan

    Examples
    --------
    >>> a = np.array([[2, 2, 2], [2, np.nan, 2], [2, 2, 2]])
    >>> b = interp_DW(a)
    """
    m = array2d.shape[0]
    n = array2d.shape[1]
    for i in range(m):
        for j in range(n):
            if np.isnan(array2d[i][j]):
                i_min = np.nan
                i_max = np.nan
                j_min = np.nan
                j_max = np.nan
                for p in range(i, -1, -1):
                    if not np.isnan(array2d[p][j]):
                        i_min = p
                        break
                for p in range(i, m, 1):
                    if not np.isnan(array2d[p][j]):
                        i_max = p
                        break
                for p in range(j, -1, -1):
                    if not np.isnan(array2d[i][p]):
                        j_min = p
                        break
                for p in range(j, n, 1):
                    if not np.isnan(array2d[i][p]):
                        j_max = p
                        break
                if np.isnan(i_min):
                    i_min = 0
                if np.isnan(i_max):
                    i_max = m - 1
                if np.isnan(j_min):
                    j_min = 0
                if np.isnan(j_max):
                    j_max = n - 1
                dis = list()
                value = list()
                for p in range(i_min, i_max + 1, 1):
                    for q in range(j_min, j_max + 1, 1):
                        if not np.isnan(array2d[p][q]):
                            dis.append(
                                1.0 / (np.abs(p - i)**2 + np.abs(q - j)**2))
                            value.append(array2d[p][q])
                interp = 0
                # for r in range(len(dis)):
                for r, _ in enumerate(dis):
                    interp = interp + dis[r] * value[r]
                try:
                    interp = interp / np.sum(dis)
                except Exception as e:
                    print(e)
                    break
                array2d[i][j] = float("%.2f" % interp)


def spline_1d(twt, vel, step, startTwt=None, endTwt=None, method='cubic'):
    startTwt = twt[0] if startTwt is None else startTwt
    endTwt = twt[-1] if endTwt is None else endTwt

    newTwt = np.arange(startTwt, endTwt, step)
    f = interpolate.interp1d(twt, vel, kind=method)
    newVel = f(newTwt)

    return (list(newTwt), list(newVel))



# -*- coding: utf-8 -*-
"""
Routines for eaton pressure prediction

Created on Sep 20 2018
"""
from __future__ import (absolute_import, division, print_function,
                        unicode_literals)

__author__ = "yuhao"

import numpy as np


def eaton(v, vn, hydrostatic, lithostatic, n=3):
    """
    Compute pore pressure using Eaton equation.

    Parameters
    ----------
    v : 1-d ndarray
        velocity array whose unit is m/s.
    vn : 1-d ndarray
        normal velocity array whose unit is m/s.
    hydrostatic : 1-d ndarray
        hydrostatic pressure in mPa
    lithostatic : 1-d ndarray
        Overburden pressure whose unit is mPa.
    v0 : float, optional
        the velocity of unconsolidated regolith whose unit is ft/s.
    n : float, optional
        eaton exponent

    Returns
    -------
    ndarray

    Notes
    -----
    .. math:: P = S - {\\sigma}_{n}\\left(\\frac{V}{V_{n}}\\right)^{n}

    [4]_

    .. [4] Eaton, B. A., & others. (1975). The equation for geopressure
       prediction from well logs. In Fall Meeting of the Society of Petroleum
       Engineers of AIME. Society of Petroleum Engineers.
    """
    ves = (lithostatic - hydrostatic) * (v / vn)**n
    pressure = lithostatic - ves
    return pressure


def sigma_eaton(es_norm, v_ratio, n):
    """
    calculate effective pressure with the ratio of velocity and normal velocity

    Notes
    -----
    .. math:: {\\sigma}={\\sigma}_{n}\\left(\\frac{V}{V_{n}}\\right)^{n}

    """
    return es_norm * (v_ratio)**n


def power_eaton(v_ratio, n):
    """
    Notes
    -----
    .. math:: \\frac{\\sigma}{{\\sigma}_{n}}=
        \\left(\\frac{V}{V_{n}}\\right)^{n}

    """
    return (v_ratio)**n



# -*- coding: utf-8 -*-
"""
A survey setting class

Created on Sat Jan 20 2018
"""
from __future__ import (absolute_import, division, print_function,
                        unicode_literals)

from builtins import range

import math

import numpy as np

from . import Path


__author__ = "Yu Hao"
__copyright__ = "Copyright (C) 2018 Yu Hao"

class SurveySetting(object):
    """
    class to hold survey settings and compute additional coordination property
    """
    def __init__(self, threepoints):
        self.threepoints = threepoints
        self.startInline = threepoints.startInline
        self.endInline = threepoints.endInline
        self.stepInline = threepoints.stepInline
        self.startCrline = threepoints.startCrline
        self.endCrline = threepoints.endCrline
        self.stepCrline = threepoints.stepCrline
        self.startDepth = threepoints.startDepth
        self.endDepth = threepoints.endDepth
        self.stepDepth = threepoints.stepDepth
        self.zType = threepoints.zType
        self.inline_A = threepoints.inline_A
        self.crline_A = threepoints.crline_A
        self.east_A = threepoints.east_A
        self.north_A = threepoints.north_A
        self.inline_B = threepoints.inline_B
        self.crline_B = threepoints.crline_B
        self.east_B = threepoints.east_B
        self.north_B = threepoints.north_B
        self.inline_C = threepoints.inline_C
        self.crline_C = threepoints.crline_C
        self.east_C = threepoints.east_C
        self.north_C = threepoints.north_C
        #--------------------
        self.inline_bin = None
        self.crline_bin = None
        self.area = None
        self.invertedAxis = None
        self.azimuth = None
        # if file is not None:
            # self._read_from_file(file_path)
        self._bin_size()
        self._basic()
        self._coordinate_conversion()
        self.azimuth_and_invertedAxis()

    def _basic(self):
        self.nEast = (self.endInline - self.startInline) // \
            self.stepInline + 1
        self.nNorth = (self.endCrline - self.startCrline) // \
            self.stepCrline + 1
        self.stepEast = math.sqrt(
            (self.north_C - self.north_B)**2 + (self.east_C - self.east_B)**2) / \
                ((self.inline_C - self.inline_B) / self.stepInline)
        self.stepNorth = math.sqrt(
            (self.north_B - self.north_A)**2 + (self.east_B - self.east_A)**2) / \
                ((self.crline_B - self.crline_A) / self.stepCrline)
        self.nDepth = (self.endDepth - self.startDepth) // \
            self.stepDepth + 1

    @staticmethod
    def angle(x, y):
        """
        Return angle from 0 to pi

        x : tuple
        y : tuple
        """
        dx = x[1] - x[0]
        dy = y[1] - y[1]
        arctan = math.atan(dy / dx)

        if dx > 0 and dy > 0:
            angle = arctan
        elif dx > 0 and dy < 0:
            angle = 2 * math.pi + arctan
        elif dx < 0 and dy > 0:
            angle = math.pi + arctan
        elif dx < 0 and dy < 0:
            angle = math.pi + arctan
        return angle

    def _bin_size(self):
        dist_ab = np.sqrt((self.north_B - self.north_A) *\
                          (self.north_B - self.north_A) + \
                          (self.east_B - self.east_A) * \
                          (self.east_B - self.east_A))
        dist_bc = np.sqrt((self.north_C - self.north_B) *\
                          (self.north_C - self.north_B) + \
                          (self.east_C - self.east_B) * \
                          (self.east_C - self.east_B))
        self.crline_bin = np.round(dist_ab / (self.crline_B - self.crline_A),
                                   decimals=2)
        self.inline_bin = np.round(dist_bc / (self.inline_C - self.inline_B),
                                   decimals=2)
        self.area = np.round(
            self.inline_bin * (self.endInline - self.startInline) * \
            self.crline_bin * (self.endCrline - self.startCrline) * 10**(-6),
            decimals=2)

    def _coordinate_conversion(self):
        """"
        calculate coefficients for line/coordination conversion
        """
        self.gamma_x = (self.east_B - self.east_A) / \
            (self.crline_B - self.crline_A)
        self.beta_x = (self.east_C - self.east_B) / \
            (self.inline_C - self.inline_B)
        self.alpha_x = self.east_A - \
            self.beta_x * self.inline_A - \
            self.gamma_x * self.crline_A
        self.gamma_y = (self.north_B - self.north_A) / \
            (self.crline_B - self.crline_A)
        self.beta_y = (self.north_C - self.north_B) / \
            (self.inline_C - self.inline_B)
        self.alpha_y = self.north_A - \
            self.beta_y * self.inline_A - \
            self.gamma_y * self.crline_A

    def line_2_coord(self, inline, crline):
        x = self.alpha_x + self.beta_x * inline + self.gamma_x * crline
        y = self.alpha_y + self.beta_y * inline + self.gamma_y * crline
        return (x, y)

    def coord_2_line(self, coordinate):
        x = coordinate[0]
        y = coordinate[1]
        d = np.matrix([[x - self.alpha_x],
                       [y - self.alpha_y]])
        G = np.matrix([[self.beta_x, self.gamma_x],
                       [self.beta_y, self.gamma_y]])
        m = G.I * d
        # m = m.astype(int)

        inline, crline = m[0][0], m[1][0]
        param_in = (inline - self.startInline) // self.stepInline + \
            ((inline - self.startInline) % self.stepInline) // \
            (self.stepInline / 2)
        inline = self.startInline + self.stepInline * param_in
        param_cr = (crline - self.startCrline) // self.stepCrline + \
            ((inline - self.startCrline) % self.stepCrline) // \
            (self.stepCrline)
        crline = self.startCrline + self.stepCrline * param_cr
        return (inline, crline)

    def azimuth_and_invertedAxis(self):
        """
        Determine azimuth (Crossline axis direction from Coordination North)
        and Inline axis is positive to the right (invertedAxis=False) or
        to the left (invertedAxis=True)
        """
        self.inclination = 0 # radius
        b_a_north = self.north_B - self.north_A
        b_a_east = self.east_B - self.east_A
        c_b_east = self.east_C - self.east_B
        c_b_north = self.north_C - self.north_B
        # crline axis in quadrant I
        if b_a_north > 0 and b_a_east > 0:
            self.azimuth = math.atan(b_a_east / b_a_north)
            if c_b_east > 0:
                self.invertedAxis = False
            else:
                self.invertedAxis = True
        # crline axis in quadrant IV
        elif b_a_north < 0 and b_a_east > 0:
            self.azimuth = -math.atan(b_a_east / (-b_a_north)) + math.pi
            if c_b_east > 0:
                self.invertedAxis = True
            else:
                self.invertedAxis = False
        # crline axis in quadrant II
        elif b_a_north > 0 and b_a_east < 0:
            self.azimuth = -math.atan(-b_a_east / b_a_north) + 2 * math.pi
            if c_b_east > 0:
                self.invertedAxis = False
            else:
                self.invertedAxis = True
        # crline axis in quadrant III
        elif b_a_north < 0 and b_a_east < 0:
            self.azimuth = math.atan(b_a_east / (-b_a_north)) + math.pi
            if c_b_east > 0:
                self.invertedAxis = True
            else:
                self.invertedAxis = False
        # crline axis on positive x
        elif b_a_north == 0 and b_a_east > 0:
            self.azimuth = 0.5 * math.pi
            if c_b_north > 0:
                self.invertedAxis = True
            else:
                self.invertedAxis = False
        # crline axis on negtive x
        elif b_a_north == 0 and b_a_east < 0:
            self.azimuth = 1.5 * math.pi
            if c_b_north < 0:
                self.invertedAxis = True
            else:
                self.invertedAxis = False
        # crline axis on positive y
        elif b_a_north > 0 and b_a_east == 0:
            self.azimuth = 0
            if c_b_east < 0:
                self.invertedAxis = True
            else:
                self.invertedAxis = False
        # crline axis on negative y
        elif b_a_north < 0 and b_a_east == 0:
            self.azimuth = math.pi
            if c_b_east > 0:
                self.invertedAxis = True
            else:
                self.invertedAxis = False

        self.azimuth = self.azimuth / math.pi * 180  # turn into degree

    def four_corner_on_canvas(self, canvas_width, canvas_height, scale_factor=0.8):
        """
        get the coordinaiton of four corners of survey area on canvas
        """
        # get four corner xy coordination
        inlines = np.array([self.startInline, self.startInline,
                            self.endInline, self.endInline])
        crlines = np.array([self.startCrline, self.endCrline,
                            self.endCrline, self.startCrline])
        x_coord, y_coord = self.line_2_coord(inlines, crlines)
        # find the minimum and maximum value of x and y
        x_sorted = np.sort(x_coord)
        y_sorted = np.sort(y_coord)
        x_min = x_sorted[0]
        x_max = x_sorted[-1]
        y_min = y_sorted[0]
        y_max = y_sorted[-1]
        # shift x y to zero
        x_coord = x_coord - x_min
        y_coord = y_coord - y_min
        # find the longer one in width and height
        width = x_max - x_min
        height = y_max - y_min
        max_dimen = max(width, height)
        # calculate shift value
        shift_x = canvas_width * (1-scale_factor)/2
        shift_y = canvas_height * (1-scale_factor)/2
        # shrink
        canvas_width *= scale_factor
        canvas_height *= scale_factor
        canvas_min_dimen = min(canvas_width, canvas_height)
        to_canvas_ratio = canvas_min_dimen / max_dimen
        x_coord = x_coord * to_canvas_ratio
        y_coord = y_coord * to_canvas_ratio
        # mirror y because origin of canvas is the upper left corner
        y_coord = canvas_height - y_coord
        # shift
        x_coord += shift_x
        y_coord += shift_y
        return x_coord, y_coord

    def draw_survey_line(self, ax):
        func = self.line_2_coord
        min_inline = self.startInline
        max_inline = self.endInline
        step_inline = self.stepInline
        if min_inline > max_inline:
            min_inline, max_inline = max_inline, min_inline
            step_inline = -step_inline

        min_crline = self.startCrline
        max_crline = self.endCrline
        step_crline = self.stepCrline
        if min_crline > max_crline:
            min_crline, max_crline = max_crline, min_crline
            step_crline = -step_crline

        for inl in range(min_inline, max_inline, step_inline):

            x_1, y_1 = func(inl, max_crline)
            x_2, y_2 = func(inl, min_crline)
            ax.plot([x_1, x_2],
                    [y_1, y_2], 'r', alpha=0.5, linewidth=0.1)

        for crl in range(min_crline, max_crline, step_crline):
            x_1, y_1 = func(min_inline, crl)
            x_2, y_2 = func(max_inline, crl)
            ax.plot([x_1, x_2],
                    [y_1, y_2], 'r', alpha=0.5, linewidth=0.1)
        ax.set(xlabel="X (Longitude)", ylabel="Y (Latitude)")

        x, y = func(min_inline, min_crline)
        ax.text(x, y, "{}/{}".format(min_inline, min_crline), color='blue')
        x, y = func(min_inline, max_crline)
        ax.text(x, y, "{}/{}".format(min_inline, max_crline), color='blue')
        x, y = func(max_inline, max_crline)
        ax.text(x, y, "{}/{}".format(max_inline, max_crline), color='blue')
        x, y = func(max_inline, min_crline)
        ax.text(x, y, "{}/{}".format(max_inline, min_crline), color='blue')



# -*- coding: utf-8 -*-
"""
a Well class utilizing pandas DataFrame and hdf5 storage

Created on Tue Dec 27 2016
"""
from __future__ import (absolute_import, division, print_function,
                        unicode_literals)

from builtins import zip, open, bytes, str

__author__ = "yuhao"

import json
from collections import OrderedDict

import numpy as np
import pandas as pd

from pygeopressure.pressure.hydrostatic import hydrostatic_pressure
from pygeopressure.pressure.eaton import eaton
from pygeopressure.pressure.bowers import bowers_varu
from pygeopressure.pressure.multivariate import pressure_multivariate
from pygeopressure.velocity.extrapolate import normal
from .well_log import Log
from .well_storage import WellStorage


class Well(object):
    """
    A class representing a well with information and log curve data.
    """
    def __init__(self, json_file, hdf_path=None):
        """
        Parameters
        ----------
        json_file : str
            path to parameter file
        hdf_path : str, optional
            path to hdf5 file used to override the one written in json_file
        """
        self.json_file = json_file
        self.hdf_file = None
        self.well_name = None
        self.loc = None
        self.kelly_bushing = None
        self.water_depth = None
        self.total_depth = None
        # self.trajectory = None
        self.data_frame = None
        self.params = None
        self.in_hdf = False
        self._parse_json()
        if hdf_path:
            self.hdf_file = hdf_path
        self._read_hdf()

    def __str__(self):
        return "Well-{}".format(self.well_name)

    def _parse_json(self):
        try:
            with open(self.json_file) as fin:
                self.params = json.load(fin, object_pairs_hook=OrderedDict)
                self.hdf_file = self.params['hdf_file']
                self.well_name = self.params['well_name']
                self.loc = self.params['loc']
                self.kelly_bushing = self.params['KB']
                self.water_depth = self.params['WD']
                self.total_depth = self.params['TD']
        except KeyError as inst:
            print(inst)

    def _read_hdf(self):
        try:
            storage = WellStorage(self.hdf_file)
            self.data_frame = storage.get_well_data(
                self.well_name.lower().replace('-', '_'))
            self.in_hdf = True
        except Exception as inst:
            print(inst)

    @property
    def depth(self):
        """
        depth values of the well

        Returns
        -------
        numpy.ndarray
        """
        if self.data_frame is not None:
            return np.around(self.data_frame['Depth(m)'].values, decimals=1)
        else:
            raise Exception("No dataframe found.")

    @property
    def logs(self):
        """
        logs stored in this well

        Returns
        -------
        list
        """
        if self.data_frame is not None:
            temp = [item.strip(')').split('(')[0] \
                for item in self.data_frame.keys()]
            temp.remove('Depth')
            return temp
        else:
            return []

    @property
    def unit_dict(self):
        """
        properties and their units
        """
        temp_dict = {
            item.strip(')').split('(')[0]: item.strip(')').split('(')[-1] \
            for item in self.data_frame.keys()}
        return {key: '' if temp_dict[key] == key else temp_dict[key] \
            for key in temp_dict.keys()}

    @property
    def hydrostatic(self):
        """
        Hydrostatic Pressure

        Returns
        -------
        numpy.ndarray
        """
        try:
            # temp_log = self.get_log('Overburden_Pressure')
            return hydrostatic_pressure(
                self.depth,
                kelly_bushing=self.kelly_bushing,
                depth_w=self.water_depth)
        except Exception as ex:
            print(ex.message)
            # print("No 'Overburden_Pressure' log found.")

    def hydro_log(self):
        """
        Returns
        -------
        Log
            Hydrostatic Pressure
        """
        hydro_log = Log()
        hydro_log.depth = np.array(self.depth)
        hydro_log.data = self.hydrostatic
        hydro_log.name = '{}_hydro'.format(self.well_name)
        hydro_log.descr = "Hydrostatic_Pressure"
        hydro_log.units = "MPa"

        return hydro_log

    @property
    def lithostatic(self):
        """
        Overburden Pressure (Lithostatic)

        Returns
        -------
        numpy.ndarray
        """
        try:
            temp_log = self.get_log('Overburden_Pressure')
            return np.array(temp_log.data)
        except KeyError:
            print("No 'Overburden_Pressure' log found.")

    @property
    def normal_velocity(self):
        """
        Normal Velocity calculated using NCT stored in well

        Returns
        -------
        numpy.ndarray
        """
        try:
            a = self.params['nct']['a']
            b = self.params['nct']['b']
            # temp_log = self.get_log('Overburden_Pressure')
            return normal(x=self.depth, a=a, b=b)
        except KeyError:
            print("No 'Overburden_Pressure' log found.")

    def get_log(self, logs, ref=None):
        """
        Retreive one or several logs in well

        Parameters
        ----------
        logs : str or list str
            names of logs to be retrieved
        ref : {'sea', 'kb'}
            depth reference, 'sea' references to sea level, 'kb' references
            to Kelly Bushing

        Returns
        -------
        Log
            one or a list of Log objects
        """
        log_list = list()
        output_list = list()
        if isinstance(logs, (bytes, str)):
            log_list.append(logs)
        elif isinstance(logs, list):
            log_list = logs
        for name in log_list:
            new_log = Log()
            new_log.name = name.lower()[:3] + '_' + \
                self.well_name.lower().replace('-', '_')
            new_log.units = self.unit_dict[name]
            new_log.descr = name
            new_log.depth = np.array(self.data_frame['Depth(m)'].values)
            new_log.data = np.array(self.data_frame[
                '{}({})'.format(name, self.unit_dict[name])].values)
            if ref == 'sea':
                shift = int(self.kelly_bushing // 0.1)
                shift_data = np.full_like(new_log.data, np.nan, dtype=np.double)
                shift_data[:-shift] = new_log.data[shift:]
                new_log.data = shift_data
            output_list.append(new_log)
        if isinstance(logs, (bytes, str)):
            return output_list[0]
        else:
            return output_list

    def add_log(self, log, name=None, unit=None):
        """
        Add new Log to current well

        Parameters
        ----------
        log : Log
            log to be added
        name : str, optional
            name for the newly added log, None, use log.name
        unit : str, optional
            unit for the newly added log, None, use log.unit
        """
        log_name = log.descr.replace(' ', '_')
        log_unit = log.units
        if name is not None:
            log_name = name
        if unit is not None:
            log_unit = unit
        if log_name not in self.logs:
            temp_dataframe = pd.DataFrame(
                data={
                    'Depth(m)':log.depth,
                    '{}({})'.format(
                        log_name, log_unit): log.data})
            self.data_frame = self.data_frame.join(
                temp_dataframe.set_index("Depth(m)"), on="Depth(m)")
        else:
            raise Warning("{} already exists in well {}".format(
                log_name, self.well_name))

    def drop_log(self, log_name):
        """
        delete a Log in current Well

        Parameters
        ----------
        log_name : str
            name of the log to be deleted
        """
        if log_name in self.logs:
            log = self.get_log(log_name)
            col = "{}({})".format(log.descr.replace(' ', '_'), log.units)
            self.data_frame = self.data_frame.drop(col, 1)
        else:
            print("no log named {}".format(log_name))

    def rename_log(self, log_name, new_log_name):
        """
        Parameters
        ----------
        log_name : str
            log name to be replaced
        new_log_name : str
        """
        if log_name in self.logs:
            log = self.get_log(log_name)
            old_str = "{}({})".format(log.descr.replace(' ', '_'), log.units)
            new_str = "{}({})".format(new_log_name.replace(' ', '_'), log.units)
            self.data_frame = self.data_frame.rename(index=str,
                                                     columns={old_str: new_str})

    def update_log(self, log_name, log):
        """
        Update well log already in current well with a new Log

        Parameters
        ----------
        log_name : str
            name of the log to be replaced in current well
        log : Log
            Log to replace
        """
        old_log = self.get_log(log_name)
        if old_log.depth == log.depth:
            self.data_frame["{}({})".format(
                old_log.descr.replace(' ', '_'), old_log.units)] = log.data
        else:
            raise Warning("Mismatch")

    def to_las(self, file_path, logs_to_export=None, full_las=False,
               null_value=1e30):
        """
        Export logs to LAS or pseudo-LAS file

        Parameters
        ----------
        file_path : str
            output file path
        logs_to_export : list of str
            Log names to be exported, None export all logs
        full_las : bool
            True, export LAS header; False export only data hence psuedo-LAS
        null_value : scalar
            Null Value representation in output file.
        """
        if logs_to_export is None:
            logs_to_export = self.logs
        keys_to_export = ["Depth(m)"]
        for log_name in logs_to_export:
            log = self.get_log(log_name)
            keys_to_export.append(
                "{}({})".format(log.descr.replace(' ', '_'), log.units))
        self.data_frame.to_csv(
            file_path, sep="\t".encode('utf-8'), na_rep="{}".format(null_value),
            columns=keys_to_export, index=False)

    def save_well_logs(self):
        """
        Save current well logs to file
        """
        try:
            storage = WellStorage(self.hdf_file)
            storage.update_well(self.well_name, self.data_frame)
        except Exception as inst:
            print(inst)
    # Measured Presure ----------
    def get_pressure(self, pres_key, ref=None, hydrodynamic=0, coef=False):
        """
        Get Pressure Values or Pressure Coefficients

        Parameters
        ----------
        pres_key : str
            Pressure data name
        ref : {'sea', 'kb'}
            depth reference, 'sea' references to sea level, 'kb' references
            to Kelly Bushing
        hydrodynamic : float
            return Pressure at depth deeper than this value
        coef : bool
            True - get pressure coefficient else get pressure value

        Returns
        -------
        Log
            Log object containing Pressure or Pressure coefficients
        """
        pres_to_get = None
        try:
            pres_to_get = self.params[pres_key]
        except KeyError:
            print("{}: Cannot find {}".format(self.well_name, pres_key))
            return Log()
        hydro = hydrostatic_pressure(self.depth,
                                     kelly_bushing=self.kelly_bushing,
                                     depth_w=self.water_depth)
        depth = pres_to_get["depth"]
        coefficients = pres_to_get["coef"]
        pres = pres_to_get["data"]

        obp_depth = self.depth # obp_log.depth

        output_depth = np.array(depth)

        if coef is True:
            if not coefficients:
                # get pressure coefficients but no coefficients stored
                output_data = list()
                for dp, pr in zip(depth, pres):
                    idx = np.searchsorted(self.depth, dp)
                    output_data.append(pr / hydro[idx])
                output_data = np.array(output_data)
            else:
                output_data = np.array(coefficients)
        elif coef is False:
            if pres:
                # get pressure values and values stored
                coefficients = pres
                hydro = np.ones(hydro.shape)

            output_data = list()
            # ouput_depth = list()
            for dp, co in zip(depth, coefficients):
                idx = np.searchsorted(obp_depth, dp)
                output_data.append(hydro[idx] * co)
                # ouput_depth.append(dp)
            output_data = np.array(output_data)
        else:
            raise Exception()

        mask = output_depth > hydrodynamic
        output_depth = output_depth[mask]
        output_data = output_data[mask]

        log = Log()
        if ref == 'sea':
            log.depth = output_depth - self.kelly_bushing
        else:
            log.depth = output_depth
        log.data = np.round(output_data, 4)

        return log

    def get_pressure_normal(self):
        """
        return pressure points within normally pressured zone.

        Returns
        -------
        Log
            Log object containing normally pressured measurements
        """
        # obp_log = self.get_log("Overburden_Pressure")
        hydro = hydrostatic_pressure(self.depth,
                                     kelly_bushing=self.kelly_bushing,
                                     depth_w=self.water_depth)
        depth = self.params["MP"]
        obp_depth = self.depth
        pres_data = list()
        for dp in depth:
            idx = np.searchsorted(obp_depth, dp)
            pres_data.append(hydro[idx])

        log = Log()
        log.depth = depth
        log.data = pres_data
        return log
    # Prediction ------------
    def eaton(self, vel_log, obp_log=None, n=None, a=None, b=None):
        """
        Predict pore pressure using Eaton method

        Parameters
        ----------
        vel_log : Log
            velocity log
        obp_log : Log
            overburden pressure log
        n : scalar
            Eaton exponent

        Returns
        -------
        Log
            a Log object containing calculated pressure.
        """
        if isinstance(vel_log, (bytes, str)):
            vel_log.get_log(vel_log)
        velocity = np.array(vel_log.data)

        if obp_log is None:
            obp = self.lithostatic
        else:
            if isinstance(obp_log, (bytes, str)):
                obp_log.get_log(obp_log)
            obp = np.array(obp_log.data)

        try:
            n = self.params['n'] if n is None else n
        except KeyError:
            n = 3

        if a is None or b is None:
            a = self.params['nct']['a']
            b = self.params['nct']['b']

        # log = Log()
        # log.depth = self.depth

        pres_eaton = eaton(hydrostatic=self.hydrostatic,
                         lithostatic=obp,
                         n=n, v=velocity, vn=self.normal_velocity)
        # log.name = "pressure_eaton_{}".format(
        #     self.well_name.lower().replace('-', '_'))
        # log.descr = "Pressure_Eaton"
        # log.units = "MPa"
        # log.prop_type = "PRE"

        log = Log.from_scratch(
            self.depth, pres_eaton,
            "pressure_eaton_{}".format(self.well_name.lower().replace('-', '_')),
            descr = "Pressure_Eaton",
            units = "MPa",
            prop_type = "PRE")

        return log

    def bowers(self, vel_log, obp_log=None, a=None, b=None, u=1, vmax=4600,
               start_depth=None, buf=20, end_depth=None, end_buffer=10):
        """
        Predict pore pressure using Eaton method

        Parameters
        ----------
        vel_log : Log
            velocity log
        obp_log : Log
            overburden pressure log
        a, b, u : float
            bowers model coefficients

        Returns
        -------
        Log
            a Log object containing calculated pressure.
        """
        if isinstance(vel_log, (bytes, str)):
            vel_log.get_log(vel_log)
        velocity = np.array(vel_log.data)

        if obp_log is None:
            obp = self.lithostatic
        else:
            if isinstance(obp_log, (bytes, str)):
                obp_log.get_log(obp_log)
            obp = np.array(obp_log.data)

        try:
            a = self.params['bowers']['A'] if a is None else a
            b = self.params['bowers']['B'] if b is None else b
            if not start_depth:
                start_depth = self.params['bowers']["start_depth"]
            if not end_depth:
                end_depth = self.params['bowers']["end_depth"]
            vmax = self.params['bowers']["vmax"] if vmax is None else vmax
            u = self.params['bowers']["U"] if u is None else u
        except KeyError as e:
            raise KeyError("Missing parameter: {}".format(e.args[0]))

        log = Log()
        log.depth = self.depth

        log.data = bowers_varu(
            velocity, obp, u, start_idx=vel_log.get_depth_idx(start_depth),
            a=a, b=b, vmax=vmax, buf=buf,
            end_idx=vel_log.get_depth_idx(end_depth), end_buffer=end_buffer)

        log.name = "pressure_bowers_{}".format(
            self.well_name.lower().replace('-', '_'))
        log.descr = "Pressure_Bowers"
        log.units = "MPa"
        log.prop_type = "PRE"
        return log

    def multivariate(self, vel_log, por_log, vsh_log, obp_log=None,
                     a0=None, a1=None, a2=None, a3=None, b=None):
        if isinstance(vel_log, (bytes, str)):
            vel_log.get_log(vel_log)
        vel = np.array(vel_log.data)
        if isinstance(por_log, (bytes, str)):
            por_log.get_log(por_log)
        phi = np.array(por_log.data)
        if isinstance(vsh_log, (bytes, str)):
            vsh_log.get_log(vsh_log)
        vsh = np.array(vsh_log.data)
        if isinstance(vsh_log, (bytes, str)):
            vsh_log.get_log(vsh_log)
        vsh = np.array(vsh_log.data)

        if obp_log is None:
            obp = self.lithostatic
        else:
            if isinstance(obp_log, (bytes, str)):
                obp_log.get_log(obp_log)
            obp = np.array(obp_log.data)

        try:
            a0 = self.params['multivariate']['a0'] if a0 is None else a0
            a1 = self.params['multivariate']['a1'] if a1 is None else a1
            a2 = self.params['multivariate']['a2'] if a2 is None else a2
            a3 = self.params['multivariate']['a3'] if a3 is None else a3
            b = self.params['multivariate']['B'] if b is None else b
        except KeyError as e:
            raise KeyError("Missing parameter: {}".format(e.args[0]))

        log = Log()
        log.depth = self.depth

        log.data = pressure_multivariate(
            obp, vel, phi, vsh, a0, a1, a2, a3,
            b, 1, 4600, start_idx=3000, end_idx=None)

        log.name = "pressure_mutlivariate_{}".format(
            self.well_name.lower().replace('-', '_'))
        log.descr = "Pressure_Multivariate"
        log.units = "MPa"
        log.prop_type = "PRE"
        return log

    def plot_horizons(self, ax, color_dict=None):
        """
        Plot horizons stored in well
        """
        horizon_dict = None
        try:
            color_dict = self.params['color_dict']
        except:
            pass
        try:
            horizon_dict = self.params['horizon']
        except KeyError:
            print("no horizons")
        for key in horizon_dict.keys():
            try:
                color = color_dict[key]
            except KeyError:
                color = 'black'
            except TypeError:
                color = 'black'
            ax.axhline(y=horizon_dict[key],
                       color=color, linewidth=0.5, zorder=0)

            import matplotlib.transforms as transforms
            trans = transforms.blended_transform_factory(
                ax.transAxes, ax.transData)
            ax.text(
                s=key, x=0.8, y=horizon_dict[key],
                color=color, transform=trans, size=9)
        ax.figure.canvas.draw()

    def save_params(self):
        """
        Save edited parameters to well information file
        """
        try:
            with open(self.json_file, "w") as fl:
                json.dump(self.params, fl, indent=4)
        except KeyError as inst:
            print(inst)



# -*- coding: utf-8 -*-
"""
an interface for interacting with Las file

Created on Thu May 10 2018
"""
from __future__ import (absolute_import, division, print_function,
                        unicode_literals)

__author__ = "yuhao"

from builtins import open
from future.utils import native
import numpy as np
import pandas as pd

from pygeopressure.pressure.hydrostatic import hydrostatic_pressure
from pygeopressure.pressure.eaton import eaton
from pygeopressure.velocity.extrapolate import normal
from pygeopressure.basic.well_log import Log
from pygeopressure.basic.las_reader import LASReader

from . import Path


class LasData(object):
    """
    Class for reading LAS and pseudo-LAS file data

    null_values could be set to more values in order to deal with messy files
    """
    def __init__(self, las_file):
        self.las_file = las_file
        self.null_values = [-999.25, 1.0e30]
        self._file_type = None
        self._data_frame = None
        self._logs = None
        self._units = None
        self.depth_unit = "m"

    @property
    def file_type(self):
        if self._file_type is None:
            with open(self.las_file, "r") as fl:
                first_line = fl.readline()
                if first_line[0] == '~':
                    self._file_type = "las"
                else:
                    self._file_type = "pseudo-las"
                return self._file_type
        else:
            return self._file_type

    @property
    def data_frame(self):
        if self._data_frame is None:
            if self.file_type is "las":
                self.read_las()
            else:
                self.read_pseudo_las()
        return self._data_frame

    def read_pseudo_las(self):
        df = pd.read_csv(self.las_file, sep='\t')
        for value in self.null_values:
            df = df.replace(value, np.nan)  # replace 1e30 with np.nan
        df = df.round({df.columns[0]: 1})  # round depth to 1 decimal
        if 'Depth(M)' in df.columns.values.tolist():
            df.rename(columns={'Depth(M)': 'Depth(m)'}, inplace=True)
        self._data_frame = df
        # data.to_hdf(self.hdf5_file, well_name)

    def read_las(self):
        if Path(native(self.las_file)).exists():
            las = LASReader(native(str(Path(self.las_file))), null_subs=np.nan)
            # well_name = las.well.items['WELL'].data
            df = pd.DataFrame(
                las.data2d,
                columns=["{}({})".format(
                    las.curves.items[name].descr.replace(' ', '_'),
                    las.curves.items[name].units) \
                    for name in las.curves.names])
            if 'Depth(M)' in df.columns.values.tolist():
                df.rename(columns={'Depth(M)': 'Depth(m)'}, inplace=True)
            self._data_frame = df

    def find_logs(self):
        self._logs = []
        self._units = []
        for name in self.data_frame.columns.values:
            str_list = name.split("(")
            if str_list[0] != "Depth":
                self._logs.append(str_list[0])
                self._units.append(str_list[1][:-1])
            else:
                self.depth_unit = str_list[1][:-1]

    @property
    def logs(self):
        if self._logs is None:
            self.find_logs()
        return self._logs

    @property
    def units(self):
        if self._units is None:
            self.find_logs()
        return self._units



# -*- coding: utf-8 -*-
"""
class Horizon for accessing horizon

Created on Fri July 20 2017
"""
from __future__ import (absolute_import, division, print_function,
                        unicode_literals)

__author__ = "yuhao"

import pandas as pd


class Horizon(object):
    """
    Horizon using excel file as input

    Parameters
    ----------
    data_file : str
        path to excel data file
    """
    def __init__(self, data_file):
        self.hdf_file = None
        self.horizon_name = None
        # self.data_frame = pd.read_excel(data_file)
        self.data_frame = pd.read_csv(data_file, sep='\t')

    def __str__(self):
        return "Horizon Object: {}".format(self.horizon_name)

    def get_cdp(self, cdp):
        """
        Get value for a CDP point on the horizon.

        Parameters
        ----------
        cdp : tuple of int (inline, crossline)
        """
        inl, crl = cdp
        return self.data_frame[
            (self.data_frame.inline == inl) & \
            (self.data_frame.crline == crl)].z.values[-1]



# -*- coding: utf-8 -*-
"""
Class for defining a seismic survey

Created on Fri Dec 11 20:24:38 2015
"""
from __future__ import (absolute_import, division, print_function,
                        unicode_literals)

from builtins import str, range, open

__author__ = "yuhao"

import json
from itertools import product
from future.utils import native

import numpy as np

from .seisegy import SeiSEGY
from .well import Well
from pygeopressure.basic.horizon import Horizon
from .survey_setting import SurveySetting
from .threepoints import ThreePoints
from . import Path

class Survey(SurveySetting):
    """
    Survey object for combining seismic data and well log data.

    Parameters
    ----------
    survey_dir : str
        survey directory.

    Attributes
    ----------
    seis_json : str
        associated seismic data information file.
    well_json : str
        associated well data information file.
    wells : dict
        dictionary holding all Well objects.
    seisCube : SeisCube
        SeisCube object holding seismic data.
    inl_crl : dict
        well position in reference to seismic survey setting (inl/crl)

    Methods
    -------
    add_well(well)
        add a well to survey
    get_seis(well_name, attr, radius=0)
        get seismic data in the vicinity of a given well
    """
    def __init__(self, survey_dir):
        """
        Parameters
        ----------
        survey_dir : str
            survey directory.
        """
        self.survey_dir = Path(native(survey_dir))
        self.setting_json = None
        self._check_dir()
        super(Survey, self).__init__(ThreePoints(self.setting_json))
        self.wells = dict()
        self.seismics = dict()
        self.inl_crl = dict()
        self.horizons = dict()
        self._add_seis_wells()
        self._add_seismic()
        self._add_horizon()

    def _check_dir(self):
        survey_file = self.survey_dir / '.survey'
        if survey_file.exists():
            self.setting_json = native(str(survey_file))
        else:
            raise Exception("No survey setting file in folder!")

    def _add_seismic(self):
        seis_dir = self.survey_dir / "Seismics"
        for seis_name in get_data_files(seis_dir):
            info_file = str(self.survey_dir.absolute() / \
                "Seismics" / "{}.seis".format(seis_name))
            data_path = None
            with open(info_file, "r") as fl:
                data_path = Path(native(json.load(fl)["path"]))
            if not data_path.is_absolute() and \
                    data_path.name == str(data_path):
                data_path = self.survey_dir.absolute() / "Seismics" / data_path
            self.seismics[seis_name] = SeiSEGY.from_json(info_file,
                                                         str(data_path))

    def _add_seis_wells(self):
        well_dir = self.survey_dir / "Wellinfo"
        for well_name in get_data_files(well_dir):
            info_file = str(self.survey_dir.absolute() / \
                "Wellinfo" / "{}.well".format(well_name))
            data_path = None
            with open(info_file, "r") as fl:
                data_path = Path(native(json.load(fl)["hdf_file"]))
            if not data_path.is_absolute() and \
                    data_path.name == str(data_path):
                data_path = self.survey_dir.absolute() / "Wellinfo" / data_path
            temp = Well(info_file, data_path)
            self.wells[temp.well_name] = temp

        for well in self.wells.values():
            loc = self._tie(well)
            self.inl_crl[well.well_name] = loc

    def _add_horizon(self):
        surf_dir = self.survey_dir / "Surfaces"
        for f in surf_dir.iterdir():
            if f.is_file() and f.suffix == '.hor':
                tmp = Horizon(str(f))
                tmp.horizon_name = f.stem
                self.horizons[f.stem] = tmp

    def _tie(self, well):
        return self.coord_2_line(well.loc)

    def get_seis(self, seis_name, well_name, radius=0):
        """
        Get seismic trace data nearest to the well location.
        """
        radius = int(radius)
        if well_name in self.inl_crl.keys():
            loc = self.inl_crl[well_name]
            if radius == 0:
                data = self.seismics[seis_name].cdp(loc)
                return [loc], [data]
            else:
                inlines, crlines = self._get_traces(seis_name, radius, loc)
                loc = list()
                data = list()
                for inl, crl in product(inlines, crlines):
                    loc.append((inl, crl))
                    data.append(self.seismics[seis_name].cdp((inl, crl)))
                return loc, data
        else:
            print("Well not found!")
            return []

    def _get_traces(self, seis_name, radius, cdp):
        inline, crline = cdp
        start_inline = self.seismics[seis_name].startInline
        end_inline = self.seismics[seis_name].endInline
        step_inline = self.seismics[seis_name].stepInline
        start_crline = self.seismics[seis_name].startCrline
        end_crline = self.seismics[seis_name].endCrline
        step_crline = self.seismics[seis_name].stepCrline

        inl_min = inline - radius * step_inline
        if inl_min < start_inline:
            inl_min = start_inline
        inl_max = inline + radius * step_inline
        if inl_max > end_inline:
            inl_max = end_inline
        crl_min = crline - radius * step_crline
        if crl_min < start_crline:
            crl_min = start_crline
        crl_max = crline + radius * step_crline
        if crl_max > end_crline:
            crl_max = end_crline
        inlines = np.arange(inl_min, inl_max+1, step_inline)
        crlines = np.arange(crl_min, crl_max+1, step_crline)
        return (inlines, crlines)

    def sparse_mesh(self, seis_name, depth, log_name):
        depth_range = range(
            self.seismics[seis_name].startDepth,
            (self.seismics[seis_name].endDepth + 1),
            self.seismics[seis_name].stepDepth)
        if depth > self.seismics[seis_name].endDepth:
            print("input depth larger than maximum depth")
            # depth = self.seismics[seis_name].endDepth
        elif depth < self.seismics[seis_name].startDepth:
            print("input depth smaller than minimum depth")
            # depth = self.seismics[seis_name].startDepth
        elif depth not in depth_range:
            print("return values on nearest depth")
        else:
            pass
        depth = min(depth_range, key=lambda x: abs(x-depth))
        mesh = np.array([np.nan] * self.seismics[seis_name].nNorth * \
            self.seismics[seis_name].nEast)
        mesh.shape = (self.seismics[seis_name].nNorth,
                      self.seismics[seis_name].nEast)
        for we in self.wells.values():
            log_depth = we.get_log("depth")
            log_data = we.get_log(log_name)
            log_index = range(len(log_depth))
            d = log_data[min(log_index, key=lambda x: abs(log_depth[x]-depth))]
            # print("d = ", d)
            w_inline = self.inl_crl[we.name][0]
            w_crline = self.inl_crl[we.name][1]
            a = (w_crline - self.seismics[seis_name].startCrline) // \
                self.seismics[seis_name].stepCrline
            b = (w_inline - self.seismics[seis_name].startInline) // \
                self.seismics[seis_name].stepInline
            mesh[a, b] = d
        return mesh

    def get_sparse_list(self, seis_name, depth, log_name):
        depth_range = range(self.seismics[seis_name].startDepth,
                            (self.seismics[seis_name].endDepth + 1),
                            self.seismics[seis_name].stepDepth)
        if depth > self.seismics[seis_name].endDepth:
            print("input depth larger than maximum depth")
            # depth = self.seismics[seis_name].endDepth
        elif depth < self.seismics[seis_name].startDepth:
            print("input depth smaller than minimum depth")
            # depth = self.seismics[seis_name].startDepth
        elif depth not in depth_range:
            print("return values on nearest depth")
        else:
            pass
        depth = min(depth_range, key=lambda x: abs(x-depth))
        sparse_list = list()
        for we in self.wells.values():
            log_depth = we.get_log("depth")
            log_data = we.get_log(log_name)
            log_index = range(len(log_depth))
            d = log_data[min(log_index, key=lambda x: abs(log_depth[x]-depth))]

            w_inline = self.inl_crl[we.name][0]
            w_crline = self.inl_crl[we.name][1]
            # a = (w_crline - self.seismics[seis_name].startCrline) // \
            #     self.seismics[seis_name].stepCrline
            # b = (w_inline - self.seismics[seis_name].startInline) // \
            #     self.seismics[seis_name].stepInline
            sparse_list.append([w_inline, w_crline, d])

        return sparse_list

# -----------------------------------------------------------------------------
# Utilities
# -----------------------------------------------------------------------------
def create_survey_directory(root_dir, survey_name):
    """
    Create survey folder structure

    Parameters
    ----------
    root_dir : str
        Root directory for storing surveys
    survey_nam : str
    """
    root_dir = Path(root_dir)
    survey_root = root_dir / survey_name
    try:
        survey_root.mkdir()
        dir_to_create = [root_dir / survey_name / 'Seismics',
                         root_dir / survey_name / 'Wellinfo',
                         root_dir / survey_name / 'Surfaces']
        for directory in dir_to_create:
            directory.mkdir()
        (survey_root / ".survey").touch()
        return survey_root
    except OSError:
        raise DuplicateSurveyNameExeption()


class DuplicateSurveyNameExeption(Exception):
    def __init__(self):
        self.message = ""
        super(DuplicateSurveyNameExeption, self).__init__(self.message)


def get_data_files(dir_path):
    """
    get all dot file with given path

    dir_path: Path
    """
    fnames = list()
    # dir_path = Path(dir_path)
    if dir_path.exists():
        # if list(dir_path.glob('.*')):
        #     for item in list(dir_path.glob('.*')):
        #         fnames.append(item.name[1:])
        for f in dir_path.iterdir():
            if f.suffix == '.seis' or f.suffix == '.well':
                fnames.append(f.stem)
    return fnames



# -*- coding: utf-8 -*-
"""
Created on Feb. 14th 2018
"""
from __future__ import (absolute_import, division, print_function,
                        unicode_literals)

from builtins import open, bytes, str

__author__ = "yuhao"

import json


class ThreePoints(object):
    """
    inline, crossline and z coordinates of three points in survey
    """
    def __init__(self, json_file=None):
        self.json_file = None
        self.dict_survey = None
        self.startInline = None
        self.endInline = None
        self.stepInline = None
        self.startCrline = None
        self.endCrline = None
        self.stepCrline = None
        self.startDepth = None
        self.endDepth = None
        self.stepDepth = None
        self.zType = None
        self.inline_A = None
        self.crline_A = None
        self.east_A = None
        self.north_A = None
        self.inline_B = None
        self.crline_B = None
        self.east_B = None
        self.north_B = None
        self.inline_C = None
        self.crline_C = None
        self.east_C = None
        self.north_C = None
        if json_file is not None:
            if isinstance(json_file, (bytes, str)):
                self.json_file = json_file
                self._read_json()
            elif isinstance(json_file, dict):
                self.dict_survey = json_file
            self._parse_survey_setting()
        else:
            raise Exception("json_file is None")

    def _read_json(self):
        try:
            with open(self.json_file, 'r') as fl:
                self.dict_survey = json.load(fl)
        except ValueError as e:
            print(e.message)

    def _parse_survey_setting(self):
        try:
            self._parse_survey_setting_v1()
        except Not_threepoints_v1_Exception:
            try:
                self._parse_survey_setting_v2()
            except Not_threepoints_v2_Exception:
                print("cannot parse file")
                raise Invalid_threepoints_Exception

    def _parse_survey_setting_v1(self):
        # with open(json_file, 'r') as file:
        #     dict_survey = json.load(file)
        try:
            coordinate = self.dict_survey["Coordinate"]
            self.inline_A, self.crline_A, self.east_A, self.north_A = \
                coordinate[0]
            self.inline_B, self.crline_B, self.east_B, self.north_B = \
                coordinate[1]
            self.inline_C, self.crline_C, self.east_C, self.north_C = \
                coordinate[2]
            inline_range = self.dict_survey["inline"]
            self.startInline, self.endInline, self.stepInline = inline_range
            crline_range = self.dict_survey["crline"]
            self.startCrline, self.endCrline, self.stepCrline = crline_range
            z_range = self.dict_survey["depth"]
            self.startDepth, self.endDepth, self.stepDepth = z_range
        except KeyError:
            raise Not_threepoints_v1_Exception
            # print(e.message)


    def _parse_survey_setting_v2(self):
        # with open(json_file, 'r') as file:
        #     dict_survey = json.load(file)
        try:
            self.startInline, self.endInline, self.stepInline = \
                self.dict_survey["inline_range"]
            self.startCrline, self.endCrline, self.stepCrline = \
                self.dict_survey["crline_range"]
            self.startDepth, self.endDepth, self.stepDepth, self.zType = \
                self.dict_survey["z_range"]
            # self.zType = self.dict_survey["z_range"][2]
            self.inline_A, self.crline_A, self.east_A, self.north_A = \
                self.dict_survey["point_A"]
            self.inline_B, self.crline_B, self.east_B, self.north_B = \
                self.dict_survey["point_B"]
            self.inline_C, self.crline_C, self.east_C, self.north_C = \
                self.dict_survey["point_C"]
        except KeyError:
            # print(e.message)
            raise Not_threepoints_v2_Exception

class Not_threepoints_v1_Exception(Exception):
    def __init__(self, message=None):
        self.message = message
        self.message = "Not three points V1 file"
        super(Not_threepoints_v1_Exception, self).__init__(self.message)


class Not_threepoints_v2_Exception(Exception):
    def __init__(self, message=None):
        self.message = message
        self.message = "Not three points V2 file"
        super(Not_threepoints_v2_Exception, self).__init__(self.message)


class Invalid_threepoints_Exception(Exception):
    def __init__(self, message=None):
        self.message = message
        self.message = "Not valid three points file"
        super(Invalid_threepoints_Exception, self).__init__(self.message)



# -*- coding: utf-8 -*-
"""
Routines to calculate pore pressure
"""
from __future__ import (absolute_import, division, print_function,
                        unicode_literals)

__author__ = "yuhao"

import numpy as np


def bowers(v, obp, u, start_idx, a, b, vmax, end_idx=None):
    """
    Compute pressure using Bowers equation.

    Parameters
    ----------
    v : 1-d ndarray
        velocity array whose unit is m/s.
    obp : 1-d ndarray
        Overburden pressure whose unit is Pa.
    v0 : float, optional
        the velocity of unconsolidated regolith whose unit is m/s.
    a : float, optional
        coefficient a
    b : float, optional
        coefficient b

    Notes
    -----
    .. math:: P = S - \\left[\\frac{(V-V_{0})}{a}\\right]^{\\frac{1}{b}}

    [3]_

    .. [3] Bowers, G. L. (1994). Pore pressure estimation from velocity data:
       accounting from overpressure mechanisms besides undercompaction:
       Proceedings of the IADC/SPE drilling conference, Dallas, 1994,
       (IADC/SPE), 1994, pp 515–530. In International Journal of Rock
       Mechanics and Mining Sciences & Geomechanics Abstracts (Vol. 31,
       p. 276). Pergamon.
    """
    sigma_max = ((vmax-1524)/a)**(1/b)
    ves = ((v - 1524) / a)**(1.0 / b)
    ves_fe = sigma_max*(((v-1524)/a)**(1/b)/sigma_max)**u
    ves[start_idx: end_idx] = ves_fe[start_idx: end_idx]
    return obp - ves


def bowers_varu(v, obp, u, start_idx, a, b, vmax, buf=20,
                end_idx=None, end_buffer=10):
    """
    Bowers Method with buffer zone above unloading zone

    Parameters
    ----------
    v : 1-d ndarray
        velocity array whose unit is m/s.
    obp : 1-d ndarray
        Overburden pressure whose unit is Pa.
    u : float
        coefficient u
    start_idx : int
        index of start of fluid expansion
    a : float, optional
        coefficient a
    b : float, optional
        coefficient b
    vmax : float
    buf : int, optional
        len of buffer interval, buf should be smaller than start_idx
    end_idx : int
        end of fluid expasion
    end_buffer : int
        len of end buffer interval
    """
    u_array = np.ones(v.shape)
    u_array[start_idx: end_idx] = u
    # start buffer
    u_buffer = np.linspace(1, u, buf)
    u_array[start_idx-buf+1: start_idx + 1] = u_buffer
    # end buffer
    if end_idx is not None:
        u_array[end_idx: end_idx + end_buffer] = np.linspace(u, 1, end_buffer)
    sigma_max = ((vmax-1524)/a)**(1/b)
    ves = sigma_max*(((v-1524)/a)**(1/b)/sigma_max)**u_array
    return obp - ves


def virgin_curve(sigma, a, b):
    "Virgin curve in Bowers' method."
    v0 = 1524
    return v0 + a * sigma**b


def invert_virgin(v, a, b):
    "invert of virgin curve."
    v0 = 1524
    return ((v-v0)/a)**(1/b)


def unloading_curve(sigma, a, b, u, v_max):
    "Unloading curve in Bowers's method."
    sigma_max = ((v_max-1524)/a)**(1/b)
    independent = sigma_max*(sigma/sigma_max)**(1/u)
    return virgin_curve(independent, a, b)


def invert_unloading(v, a, b, u, v_max):
    "invert of Unloading curve in Bowers's method."
    sigma_max = invert_virgin(v_max, a, b)
    sigma_vc = invert_virgin(v, a, b)
    return sigma_max * (sigma_vc/sigma_max)**u


def power_bowers(sigma_vc_ratio, u):
    return sigma_vc_ratio**u



# -*- coding: utf-8 -*-
"""
well log processing tools

Created on Sep 19 2018
"""
from __future__ import (absolute_import, division, print_function,
                        unicode_literals)

__author__ = "yuhao"

from builtins import range, open
import numpy as np
from scipy.interpolate import interp1d
from scipy.signal import butter, filtfilt

from pygeopressure.velocity.smoothing import smooth

from pygeopressure.pressure.obp import traugott_trend
from pygeopressure.basic.well_log import Log


def extrapolate_log_traugott(den_log, a, b, kb=0, wd=0):
    """
    Extrapolate density log using Traugott equation
    """
    density_trend = traugott_trend(
        np.array(den_log.depth), a, b, kb=kb, wd=wd)

    extra_log = Log()
    extra_log.name = den_log.name + "_ex"
    extra_log.units = den_log.units
    extra_log.descr = "Density_extra"
    extra_log.depth = np.array(den_log.depth)

    new_data = np.full_like(density_trend, np.nan)
    new_data[:den_log.start_idx] = density_trend[:den_log.start_idx]

    old_data = np.array(den_log.data)
    new_data[den_log.start_idx:] = old_data[den_log.start_idx:]

    extra_log.data = new_data

    return extra_log


def rolling_window(a, window):
    a = np.array(a)
    shape = a.shape[:-1] + (a.shape[-1] - window + 1, window)
    strides = a.strides + (a.strides[-1],)
    rolled = np.lib.stride_tricks.as_strided(
        a, shape=shape, strides=strides)
    return rolled


def despike(curve, curve_sm, max_clip):
    spikes = np.where(curve - curve_sm > max_clip)[0]
    spukes = np.where(curve_sm - curve > max_clip)[0]
    out = np.copy(curve)
    out[spikes] = curve_sm[spikes] + max_clip
    out[spukes] = curve_sm[spukes] - max_clip
    return out


def smooth_log(log, window=1500):
    """
    Parameters
    ----------
    log : Log object
        log to smooth
    window : scalar
        window size of the median filter

    Returns
    -------
    smoothed log : Log object
        smoothed log
    """
    data = np.array(log.data)
    depth = np.array(log.depth)
    mask = np.isfinite(data)
    func = interp1d(depth[mask], data[mask])
    interp_data = func(depth[log.start_idx: log.stop_idx])
    new_data = np.array(data)
    new_data[log.start_idx: log.stop_idx] = interp_data
    smoothed = smooth(new_data[log.start_idx: log.stop_idx], window_len=window//2, window='flat')
    # using half the window length in order to be consistent with opendtect
    smooth_data = np.array(data)
    smooth_data[log.start_idx: log.stop_idx] = smoothed
    log_smooth = Log()
    log_smooth.name = log.name + "_sm"
    log_smooth.units = log.units
    log_smooth.descr = log.descr
    log_smooth.depth = log.depth
    log_smooth.data = smooth_data
    return log_smooth


def upscale_log(log, freq=20):
    """
    downscale a well log with a lowpass butterworth filter
    """
    depth = np.array(log.depth)
    data = np.array(log.data)
    mask = np.isfinite(data)
    func = interp1d(depth[mask], data[mask])
    interp_data = func(depth[log.start_idx: log.stop_idx])
    nyq = 10000 / 2
    dw = freq / nyq
    b, a = butter(4, dw, btype='low', analog=False)
    filtered = filtfilt(b, a, interp_data, method='gust')
    downscale_data = np.array(data)
    downscale_data[log.start_idx: log.stop_idx] = filtered
    log_downscale = Log()
    log_downscale.name = log.name + "_downscale_" + str(freq)
    log_downscale.units = log.units
    log_downscale.descr = log.descr
    log_downscale.depth = log.depth
    log_downscale.data = downscale_data
    return log_downscale


def truncate_log(log, top, bottom):
    """
    Remove unreliable values in the top and bottom section of well log

    Parameters
    ----------
    log : Log object
    top, bottom : scalar
        depth value

    Returns
    -------
    trunc_log : Log object
    """
    depth = np.array(log.depth)
    data = np.array(log.data)
    if top != 0:
        mask = depth < top
        data[mask] = np.nan
    if bottom != 0:
        mask = depth > bottom
        data[mask] = np.nan
    trunc_log = Log()
    trunc_log.name = log.name + '_trunc'
    trunc_log.units = log.units
    trunc_log.descr = log.descr
    trunc_log.depth = depth
    trunc_log.data = data
    return trunc_log


def shale(log, vsh_log, thresh=0.35):
    """
    Discern shale intervals

    log : Log
        log to discern
    vsh_log : Log
        shale volume log
    thresh : scalar
        percentage threshold, 0 < thresh < 1
    """
    shale_mask = np.isfinite(vsh_log.depth)
    shale_mask[vsh_log.start_idx: vsh_log.stop_idx] = True
    mask_thresh = np.array(vsh_log.data) < thresh
    mask = shale_mask * mask_thresh
    data = np.array(log.data)
    data[mask] = np.nan
    log_sh = Log()
    log_sh.name = log.name + "_sh"
    log_sh.units = log.units
    log_sh.descr = log.descr
    log_sh.depth = log.depth
    log_sh.data = data
    return log_sh


def interpolate_log(log):
    """
    Log curve interpolation
    """
    depth = np.array(log.depth)
    data = np.array(log.data)
    mask = np.isfinite(data)
    func = interp1d(depth[mask], data[mask])
    interp_data = func(depth[log.start_idx: log.stop_idx])
    data[log.start_idx: log.stop_idx] = interp_data

    interp_log = Log()
    interp_log.name = log.name + '_interp'
    interp_log.units = log.units
    interp_log.descr = log.descr
    interp_log.depth = depth
    interp_log.data = data
    return interp_log


def local_average(log, rad=10):
    """upscale data using local averaging

    Parameters
    ----------
    data : Log()
        log data to be upscaled
    rad : int
        local radius, data within this radius will be represented by a single value

    Returns
    -------
    new_log : Log()
        upscaled log data
    """
    data = np.array(log.data)
    mask = np.isfinite(data)
    index = np.where(mask)
    start = index[0][0]
    end = index[0][-1]+1
    interval = data[start: end]
    index_toadd = []
    data_toadd = []

    step = rad*2+1
    n = len(interval)
    for i in range(0, n, step):
        seg = interval[i: i+step]
        new_mask = np.isfinite(seg)
        if len(seg[new_mask]) > rad+1:
            data_toadd.append(np.mean(seg[new_mask]))
            index_toadd.append(start+i+rad)
    data_toadd, index_toadd = np.array(data_toadd), np.array(index_toadd)
    new_data = np.full_like(data, np.nan)
    new_data[index_toadd] = data_toadd
    new_log = Log()
    new_log.depth = log.depth
    new_log.data = new_data
    return new_log


def write_peudo_las(file_name, logs):
    """
    Write multiple logs to a pseudo las file.
    """
    try:
        with open(file_name, 'w') as fout:
            description = ["Depth(m)"]
            for log in logs:
                split_list = log.descr.split(' ')
                description.append('_'.join(split_list)+"(" + log.units + ")")
            first_line = '\t'.join(description)
            fout.write(first_line + "\n")
            data = [logs[0].depth] + [log.data for log in logs]
            data = np.stack(data)
            for ds in np.nditer(data, flags=['external_loop'], order='F'):
                line = [str(v) if np.isfinite(v) else "1e30" for v in ds]
                fout.write("\t".join(line) + "\n")
            return True
    except Exception as inst:
        print(inst.args)



# -*- coding: utf-8 -*-
"""
Functions related to density and Overburden Pressrue Calculation
"""
from __future__ import (absolute_import, division, print_function,
                        unicode_literals)

__author__ = "yuhao"

from collections import OrderedDict
import numpy as np
from pygeopressure.basic.well_log import Log
from pygeopressure.basic.indexes import InlineIndex
from pygeopressure.pressure.utils import create_seis, create_seis_info


def traugott(z, a, b):
    r"""
    estimate density with depth

    Parameters
    ----------
    depth : 1-d ndarray
    a, b: scalar

    Notes
    -----
    .. math:: \overline{\rho (h)}=16.3+{h/3125}^{0.6}

    gives the average sediment density in pounds per gallon (ppg) mud weight
    equivalent between the sea floor and depth h (in feet) below the sea floor.

    So, density variation with depth takes the form [2]_:

    .. math:: \rho(z) = {\rho}_{0} + a{z}^{b}

    .. [2] Traugott, Martin. "Pore/fracture pressure determinations in deep
       water." World Oil 218.8 (1997): 68-70.
    """
    # rho0 = 2.65
    return 1.70 + a * z**b


def traugott_trend(depth, a, b, kb=0, wd=0):
    depth_shift = np.array(depth) - kb - wd
    density = traugott(depth_shift, a, b)
    mask = depth_shift < 0
    density[mask] = np.nan
    return density


def gardner(v, c=0.31, d=0.25):
    r"""
    Estimate density with velocity

    Parameters
    ----------
    v : 1-d ndarray
        interval velocity array
    c : float, optional
        coefficient a
    d : float, optional
        coefficient d

    Returns
    -------
    out : 1-d ndarray
        density array

    Notes
    -----
    .. math:: \rho = c{V}^{d}

    typical values for a and b in GOM coast are a=0.31, b=0.25 [1]_.

    .. [1] G. Gardner, L. Gardner, and A. Gregory, "Formation velocity and density -
       the diagnostic basics for stratigraphic traps," Geophysics, vol. 39,
       no. 6, pp. 770-780, 1974.
    """
    return c * v**d


def gardner_seis(output_name, vel_cube, c=0.31, d=0.25):
    """
    Parameters
    ----------
    output_name : str
        output file name without extention

    Returns
    -------
    SeiSEGY
    """
    den_cube = create_seis(output_name, vel_cube)
    create_seis_info(den_cube, output_name)
    # calculate density
    for inl in vel_cube.inlines():
        vel_inline = vel_cube.data(InlineIndex(inl))
        den_inline = gardner(vel_inline, c, d)
        den_cube.update(InlineIndex(inl), den_inline)

    return den_cube


def overburden_pressure(depth, rho, kelly_bushing=41, depth_w=82, rho_w=1.01):
    """
    Calculate Overburden Pressure

    Parameters
    ----------
    depth : 1-d ndarray
    rho : 1-d ndarray
        density in g/cm3
    kelly_bushing : scalar
        kelly bushing elevation in meter
    depth_w : scalar
        from sea level to sea bottom (a.k.a mudline) in meter
    rho_w : scalar
        density of sea water - depending on the salinity of sea water
        (1.01-1.05g/cm3)

    Returns
    -------
    obp : 1-d ndarray
        overburden pressure in mPa
    """
    g = 9.80665  # m/s2
    depth = np.array(depth)
    rho = np.array(rho)
    rho = 1000 * rho  # convert unit from g/cm3 to kg/m3
    rho_w = 1000 * rho_w
    # sea bottom to sea level
    mask = depth < (kelly_bushing + depth_w)
    rho[mask] = rho_w
    mask = depth < kelly_bushing
    rho[mask] = 0
    delta_h = np.full_like(depth, np.nan)
    delta_h[1:] = depth[1:] - depth[:-1]
    delta_h[0] = 0
    obp = rho * delta_h * g
    obp = np.cumsum(obp)
    return obp / 1000000  # mPa


def obp_well(den_log, kb=41, wd=82, rho_w=1.01):
    """
    Compute Overburden Pressure for a Log

    Parameters
    ----------
    den_log : Log
        density log (extrapolated)
    kb : scalar
        kelly bushing elevation in meter
    wd : scalar
        from sea level to sea bottom (a.k.a mudline) in meter
    rho_w : scalar
        density of sea water - depending on the salinity of sea water
        (1.01-1.05g/cm3)

    Returns
    -------
    out : Log
        Log containing overburden pressure in mPa
    """
    depth = np.array(den_log.depth)
    rho = np.array(den_log.data)
    obp = overburden_pressure(
        depth, rho, kelly_bushing=kb, depth_w=wd, rho_w=rho_w)

    obp_log = Log()
    obp_log.depth = depth
    obp_log.data = obp
    obp_log.name = "log_obp"
    obp_log.descr = "Overburden_Pressure"
    obp_log.units = "MPa"

    return obp_log


def obp_trace(rho, step):
    """
    Compute Overburden Pressure for a trace

    Parameters
    ----------
    rho : 1-d array
        density in g/cc

    Returns
    -------
    out : 1-d ndarray
        overburden pressure in mPa
    """
    data = np.array(rho)
    return np.cumsum(data * 9.8 * step * 0.001)


def obp_seis(output_name, den_cube):
    # create seismic object
    obp_cube = create_seis(output_name, den_cube)
    # create info file
    create_seis_info(obp_cube, output_name)
    # calculate obp
    step = den_cube.stepDepth
    for inl in den_cube.inlines():
        den_inline = den_cube.data(InlineIndex(inl))
        obp_inline = obp_section(den_inline, step)
        obp_cube.update(InlineIndex(inl), obp_inline)

    return obp_cube


def obp_section(rho_inline, step):
    return np.cumsum(rho_inline * 9.8 * step * 0.001, axis=1)



# -*- coding: utf-8 -*-
"""
class for interfacing with segy file.

Created on Feb. 7th 2018
"""
from __future__ import (absolute_import, division, print_function,
                        unicode_literals)

__author__ = "yuhao"

from builtins import range, open
import json
from shutil import copyfile
from itertools import product
from future.utils import native
import segyio

from .utils import  methdispatch
from .vawt import wiggles, img
from .indexes import InlineIndex, CrlineIndex, DepthIndex, CdpIndex
from .survey_setting import SurveySetting
from .threepoints import ThreePoints

from . import Path


class SeiSEGY(object):
    def __init__(self, segy_file, like=None):
        """
        Parameters
        ----------
        segy_file : str
            segy file path
        like : str, optional
            created segy file has the same dimesions as like.
        """
        self.segy_file = segy_file
        self.inDepth = False # True if dataset Z is in Depth
        self.property_type = None

        if like is not None:
            if Path(native(like)).exists() and not Path(native(self.segy_file)).exists():
                copyfile(src=like, dst=self.segy_file)

        if Path(native(self.segy_file)).exists():
            self._parse_segy()
        else:
            raise Exception("File does not exist!")

    @classmethod
    def from_json(cls, json_file, segy_file=None):
        """
        Initialize SeiSEGY from an json file containing information

        Parameters
        ----------
        json_file : str
            json file path
        segy_file : str
            segy file path for overriding information in json file.
        """
        with open(json_file, 'r') as fl:
            json_object = json.load(fl)
            segy = json_object["path"]
            inDepth = json_object["inDepth"]
            property_type = json_object["Property_Type"]

        if segy_file:
            segy = segy_file
        instance = cls(native(segy))
        instance.inDepth = inDepth
        instance.property_type = property_type

        return instance

    def __str__(self):
        return "SeiSEGY(inl[{},{},{}];crl[{},{},{}];z[{},{},{}])".format(
            self.startInline, self.endInline, self.stepInline,
            self.startCrline, self.endCrline, self.stepCrline,
            self.startDepth, self.endDepth, self.stepDepth)

    def _parse_segy(self):
        with segyio.open(self.segy_file, 'r') as segyfile:
            segyfile.mmap()
            self.startInline = segyfile.ilines[0]
            self.endInline = segyfile.ilines[-1]
            self.nEast = len(segyfile.ilines)
            self.stepInline = (self.endInline - self.startInline) // \
                (self.nEast - 1)
            self.startCrline = segyfile.xlines[0]
            self.endCrline = segyfile.xlines[-1]
            self.nNorth = len(segyfile.xlines)
            self.stepCrline = (self.endCrline - self.startCrline) // \
                (self.nNorth - 1)
            self.startDepth = segyfile.samples[0]
            self.endDepth = segyfile.samples[-1]
            self.nDepth = len(segyfile.samples)
            self.stepDepth = (self.endDepth - self.startDepth) // \
                (self.nDepth - 1)

            inline_A = self.startInline
            crline_A = self.startCrline
            index_A = 0
            x_A = segyfile.header[index_A][segyio.su.cdpx]
            y_A = segyfile.header[index_A][segyio.su.cdpy]

            inline_B = inline_A
            crline_B = self.startCrline + 2 * self.stepCrline
            index_B = 2
            x_B = segyfile.header[index_B][segyio.su.cdpx]
            y_B = segyfile.header[index_B][segyio.su.cdpy]

            inline_C = self.startInline + 2 * self.stepInline
            crline_C = crline_B
            index_C = 2 * self.nNorth + 2
            x_C = segyfile.header[index_C][segyio.su.cdpx]
            y_C = segyfile.header[index_C][segyio.su.cdpy]

            setting_dict = {
                "inline_range": [
                    self.startInline, self.endInline, self.stepInline],
                "crline_range": [
                    self.startCrline, self.endCrline, self.stepCrline],
                "z_range": [
                    self.startDepth, self.endDepth, self.stepDepth, "unknown"],
                "point_A": [inline_A, crline_A, x_A, y_A],
                "point_B": [inline_B, crline_B, x_B, y_B],
                "point_C": [inline_C, crline_C, x_C, y_C]
            }
            self.survey_setting = SurveySetting(ThreePoints(setting_dict))

    def inlines(self):
        """
        Iterator for inline numbers

        Yields
        ------
        int
            inline number
        """
        for inline in range(self.startInline, self.endInline+1, self.stepInline):
            yield inline

    def crlines(self):
        """
        Iterator for crline numbers

        Yields
        ------
        int
            cross-line number
        """
        for crline in range(self.startCrline, self.endCrline+1, self.stepCrline):
            yield crline

    def inline_crlines(self):
        """
        Iterator for both inline and crline numbers


        Yields
        ------
        tuple of int
            (inline number, crossline number)
        """
        for inline, crline in product(
                range(self.startInline, self.endInline+1, self.stepInline),
                range(self.startCrline, self.endCrline+1, self.stepCrline)):
            yield (inline, crline)

    def depths(self):
        """
        Iterator for z coordinate

        Yields
        ------
        float
            depth value
        """
        for i in range(self.nDepth):
            yield self.startDepth + i * self.stepDepth

    def inline(self, inline):
        "data of a inline section"
        with segyio.open(self.segy_file, 'r') as segyfile:
            segyfile.mmap()
            data = segyfile.iline[inline]
        return data

    def crline(self, crline):
        "data of a crossline section"
        with segyio.open(self.segy_file, 'r') as segyfile:
            segyfile.mmap()
            data = segyfile.xline[crline]
        return data

    def depth(self, depth):
        "data of a depth slice"
        depth_idx = int((depth - self.startDepth) // self.stepDepth)
        with segyio.open(self.segy_file, 'r') as segyfile:
            segyfile.mmap()
            data = segyfile.depth_slice[depth_idx]
        return data

    def cdp(self, cdp):
        "data of a cdp"
        with segyio.open(self.segy_file, 'r') as segyfile:
            segyfile.mmap()
            data = segyfile.gather[cdp]
            data = data.reshape((data.shape[-1],))
        return data

    @methdispatch
    def data(self, indexes):
        """
        Retrieve Data according to the index provided.

        Parameters
        ----------
        indexes : {InlineIndex, CrlineIndex, DepthIndex, CdpIndex}
            index of data to retrieve

        Returns
        -------
        numpy.ndarray
        """
        raise TypeError("Unsupported Type")

    @data.register(InlineIndex)
    def _(self, indexes):
        """
        data of a Inline section

        Paramaters
        ----------
        indexes : InlineIndex

        Returns
        -------
        out : 2-d ndarray
            of size nCrline * nDepth
        """
        return self.inline(indexes.value)

    @data.register(CrlineIndex)
    def _(self, indexes):
        """
        data of a Crossline section

        Paramaters
        ----------
        indexes : CrlineIndex

        Returns
        -------
        out : 2-d ndarray
            of size nInline * nDepth
        """
        return self.crline(indexes.value)

    @data.register(DepthIndex)
    def _(self, indexes):
        """
        data of a depth slice

        Paramaters
        ----------
        indexes : DepthIndex

        Returns
        -------
        out : 2-d ndarray
            of size nInline * nCrline
        """
        return self.depth(indexes.value)

    @data.register(CdpIndex)
    def _(self, indexes):
        """
        data of a cdp

        Paramaters
        ----------
        indexes : CdpIndex

        Returns
        -------
        out : 1-d ndarray
            of length nDepth
        """
        return self.cdp(indexes.value)

    def update(self, index, data):
        """
        Update data with ndarray

        Parameters
        ----------
        index : InlineIndex
        data : 2-d ndarray
            data for updating Inline
        """
        try:
            if not isinstance(index, InlineIndex):
                raise TypeError("has to be InlineIndex")
            if data.shape != (self.nNorth, self.nDepth):
                raise ValueError
            with segyio.open(self.segy_file, 'r+') as segyfile:
                segyfile.mmap()
                segyfile.iline[index.value] = data
        except Exception as er:
            print(er.message)

    @methdispatch
    def plot(self, index, ax, kind='vawt', cm='seismic', ptype='seis'):
        """
        Plot seismic section according to index provided.

        Parameters
        ----------
        index : {InlineIndex, CrlineIndex, DepthIndex, CdpIndex}
            index of data to plot
        ax : matplotlib.axes._subplots.AxesSubplot
            axis to plot on
        kind : {'vawt', 'img'}
            'vawt' for variable area wiggle trace plot
            'img' for variable density plot
        cm : str
            colormap for plotting
        ptype : str, optional
            property type

        Returns
        -------
        matplotlib.image.AxesImage
         """
        raise TypeError('Unsupported index type')

    @plot.register(InlineIndex)
    def _(self, index, ax, kind='vawt', cm='seismic', ptype='seis'):
        data = self.data(index)
        handle = None
        if kind == 'vawt':
            wiggles(data.T, wiggleInterval=1, ax=ax)
        elif kind == 'img':
            handle = img(data.T,
                         extent=[
                             self.startCrline, self.endCrline,
                             self.startDepth, self.endDepth],
                         ax=ax, cm=cm, ptype=ptype)
            ax.invert_yaxis()
        else:
            pass
        ax.get_figure().suptitle('In-line Section: {}'.format(index.value))
        from matplotlib.offsetbox import AnchoredText
        z_text = AnchoredText(
            r"$\downarrow$Z",
            loc=2, prop=dict(size=10), frameon=False,
            bbox_to_anchor=(0., 0.),
            bbox_transform=ax.transAxes)
        ax.add_artist(z_text)
        inline_text = AnchoredText(
            r"Cross-line $\rightarrow$ ",
            loc=1, prop=dict(size=10), frameon=False,
            bbox_to_anchor=(1., 0.),
            bbox_transform=ax.transAxes)
        ax.add_artist(inline_text)

        return handle

    @plot.register(CrlineIndex)
    def _(self, index, ax, kind='vawt', cm='seismic', ptype='seis'):
        data = self.data(index)
        handle = None
        if kind == 'vawt':
            wiggles(data.T, wiggleInterval=1, ax=ax)
        elif kind == 'img':
            handle = img(data.T,
                         extent=[
                             self.startInline, self.endInline,
                             self.startDepth, self.endDepth],
                         ax=ax, cm=cm, ptype=ptype)
            ax.invert_yaxis()
        else:
            pass
        ax.get_figure().suptitle('Cross-line Section: {}'.format(index.value))
        from matplotlib.offsetbox import AnchoredText
        z_text = AnchoredText(
            r"$\downarrow$Z",
            loc=2, prop=dict(size=10), frameon=False,
            bbox_to_anchor=(0., 0.),
            bbox_transform=ax.transAxes)
        ax.add_artist(z_text)
        inline_text = AnchoredText(
            r"In-line $\rightarrow$ ",
            loc=1, prop=dict(size=10), frameon=False,
            bbox_to_anchor=(1., 0.),
            bbox_transform=ax.transAxes)
        ax.add_artist(inline_text)

        return handle

    @plot.register(DepthIndex)
    def _(self, index, ax, kind='vawt', cm='seismic', ptype='seis'):
        data = self.data(index)
        handle = None
        if kind == 'vawt':
            wiggles(data.T, wiggleInterval=1, ax=ax)
        elif kind == 'img':
            handle = img(data.T,
                         extent=[
                             self.startInline, self.endInline,
                             self.startCrline, self.endCrline,],
                         ax=ax, cm=cm, ptype=ptype)
            ax.invert_yaxis()
        else:
            pass
        ax.get_figure().suptitle("Z slice: {}".format(index.value))
        from matplotlib.offsetbox import AnchoredText
        z_text = AnchoredText(
            r"$\downarrow$Cross-line",
            loc=2, prop=dict(size=10), frameon=False,
            bbox_to_anchor=(0., 0.),
            bbox_transform=ax.transAxes)
        ax.add_artist(z_text)
        inline_text = AnchoredText(
            r"In-line $\rightarrow$ ",
            loc=1, prop=dict(size=10), frameon=False,
            bbox_to_anchor=(1., 0.),
            bbox_transform=ax.transAxes)
        ax.add_artist(inline_text)

        return handle

    def valid_cdp(self, cdp_num):
        "Return valid CDP numbers nearest to cdp_num"
        inl_num, crl_num = cdp_num
        n_inline = (inl_num - self.startInline) // self.stepInline
        in_plus_one = round(((inl_num - self.startInline) % self.stepInline) / \
                            self.stepInline)
        inline = self.startInline + (n_inline + in_plus_one) * self.stepInline

        n_crline = (crl_num - self.startCrline) // self.stepCrline
        cr_plus_one = round(((crl_num - self.startCrline) % self.stepCrline) / \
                            self.stepCrline)
        crline = self.startCrline + (n_crline + cr_plus_one) * self.stepCrline
        return (inline, crline)



#!/usr/bin/env python
"""
Created on Nov 1st 2017
"""
from os import path
from distutils.core import setup
from setuptools import find_packages
import versioneer

CLASSIFIERS = [
    'Development Status :: 4 - Beta',
    'Intended Audience :: Developers',
    'Intended Audience :: Science/Research',
    'License :: OSI Approved :: MIT License',
    'Programming Language :: Python',
    'Topic :: Scientific/Engineering',
    'Topic :: Scientific/Engineering :: Mathematics',
    'Topic :: Scientific/Engineering :: Physics',
    'Operating System :: Microsoft :: Windows',
    'Operating System :: POSIX',
    'Operating System :: Unix',
    'Operating System :: MacOS',
    'Natural Language :: English',
]

FILE_DIR = path.abspath(path.dirname(__file__))
with open(path.join(FILE_DIR, 'README.md')) as f:
    LONG_DESCRIPTION = f.read()

setup(
    name="pyGeoPressure",
    version=versioneer.get_version(),
    cmdclass=versioneer.get_cmdclass(),
    python_requires=">=2.7, <3.7",
    install_requires=[
        'future',
        'pathlib2;python_version=="2.7"',
        'singledispatch;python_version=="2.7"',
        'scipy',
        'scikit-learn',
        'pandas',
        'tables',
        'matplotlib<3.0;python_version=="2.7"',
        'matplotlib;python_version>"2.7"',
        'segyio'
    ],
    packages=find_packages(exclude=['test']),
    author="Yu Hao",
    author_email="yuhao@live.cn",
    description="pyGeoPressure: Tools for geopressure prediction",
    long_description=LONG_DESCRIPTION,
    long_description_content_type='text/markdown',
    license="MIT",
    keywords="pore pressure prediction",
    url="https://github.com/whimian/pyGeoPressure",
    download_url="https://github.com/whimian/pyGeoPressure",
    classifiers=CLASSIFIERS,
    platforms=["Windows", "Linux", "Solaris", "Mac OS-X", "Unix"],
    zip_safe=False
)



# -*- coding: utf-8 -*-
"""
optimizer for different models

Created on Sep 16 2018
"""
from __future__ import (absolute_import, division, print_function,
                        unicode_literals)

__author__ = "yuhao"

from builtins import bytes, str
import json
from collections import OrderedDict

import numpy as np
from scipy.optimize import curve_fit

from pygeopressure.pressure.bowers import (
    virgin_curve, invert_virgin, power_bowers)
from pygeopressure.velocity.extrapolate import normal, normal_dt
from pygeopressure.pressure.obp import traugott
from pygeopressure.basic.well import Well
from pygeopressure.basic.well_log import Log
from pygeopressure.basic.utils import rmse, pick_sparse
from pygeopressure.pressure.eaton import power_eaton
from sklearn import linear_model


def optimize_bowers_virgin(well, vel_log, obp_log, upper, lower,
                           pres_log='loading', mode='nc', nnc=5):
    """
    Optimizer for Bowers loading curve

    Parameters
    ----------
    well : Well
    vel_log : Log or str
        Log object or well log name stored in well
    obp_log : Log or str
        Log object or well log name stored in well
    upper : float or str
        upper bound of nct, depth value or horizon name
    lower : float or str
        lower bound of nct, depth value or horizon name
    pres_log : Log or str
        Log object storing measured pressure value or Pressure name
        stored in well
    mode : {'nc', 'pres', 'both'}
        which pressure to use for optimization,
        - 'nc' : points on NCT
        - 'pres' : points in pres_log
        - 'both' : both of them
    nnc : int
        number of points to pick on NCT

    Returns
    -------
    a, b : tuple of floats
        optimized bowers loading curve coefficients
    rms_err : float
        root mean square error of pressure
    """
    if isinstance(upper, (bytes, str)):
        depth_upper = well.params['horizon'][upper]
    else:
        depth_upper = upper
    if isinstance(upper, (bytes, str)):
        depth_lower = well.params['horizon'][lower]
    else:
        depth_lower = lower
    if isinstance(vel_log, (bytes, str)):
        vel_log = well.get_log(vel_log)
    if isinstance(obp_log, (bytes, str)):
        obp_log = well.get_log(obp_log)
    if isinstance(pres_log, (bytes, str)):
        # pres_log = well.get_loading_pressure()
        pres_log = well.get_pressure(pres_log)
    depth = np.array(obp_log.depth)

    nct_vel_to_fit = []
    nct_es_to_fit = []
    pres_vel_to_fit = []
    pres_es_to_fit = []
    if mode == 'nct' or mode == 'both':
        nct_es_data = np.array(obp_log.data) - np.array(well.hydrostatic)
        nct_mask = depth < depth_lower
        nct_mask *= depth > depth_upper
        nct_mask *= depth < vel_log.stop

        nct_vel_interval = np.array(vel_log.data)[nct_mask]
        nct_es_interval = nct_es_data[nct_mask]

        nct_vel_to_fit = np.array(pick_sparse(nct_vel_interval, nnc))
        nct_es_to_fit = np.array(pick_sparse(nct_es_interval, nnc))
    if mode == 'pres' or mode == 'both':
        vel = list()
        obp = list()
        pres = list()
        for dp in pres_log.depth:
            idx = np.searchsorted(depth, dp)
            vel.append(vel_log.data[idx])
            obp.append(obp_log.data[idx])
        vel, obp, pres = np.array(vel), np.array(obp), np.array(pres_log.data)
        es = obp - pres

        pres_vel_to_fit = vel
        pres_es_to_fit = es
    vel_to_fit = np.append(nct_vel_to_fit, pres_vel_to_fit)
    es_to_fit = np.append(nct_es_to_fit, pres_es_to_fit)

    popt, _ = curve_fit(virgin_curve, es_to_fit, vel_to_fit)
    a, b = popt

    es_predicted = invert_virgin(vel_to_fit, a, b)
    rms_err = rmse(es_to_fit, es_predicted)

    return a, b, rms_err


def optimize_bowers_unloading(well, vel_log, obp_log, a, b,
                              vmax, pres_log='unloading'):
    """
    Optimize for Bowers Unloading curve parameter U

    Parameters
    ----------
    well : Well
    vel_log : Log or str
        Log object or well log name stored in well
    obp_log : Log or str
        Log object or well log name stored in well
    vmax : float
        vmax in bowers unloading curve
    pres_log : Log or str
        Log object storing measured pressure value or Pressure name
        stored in well

    Returns
    -------
    u : float
        unloading curve cofficient U
    error : float
        Relative RMS error
    """
    if isinstance(vel_log, (bytes, str)):
        vel_log = well.get_log(vel_log)
    if isinstance(obp_log, (bytes, str)):
        obp_log = well.get_log(obp_log)
    if isinstance(pres_log, (bytes, str)):
        pres_log = well.get_pressure(pres_log)

    depth = np.array(obp_log.depth)

    vel = list()
    obp = list()
    pres = list()
    for dp in pres_log.depth:
        idx = np.searchsorted(depth, dp)
        vel.append(vel_log.data[idx])
        obp.append(obp_log.data[idx])
    vel, obp, pres = np.array(vel), np.array(obp), np.array(pres_log.data)
    es = obp - pres

    sigma_max = invert_virgin(vmax, a, b)

    sigma_vc = invert_virgin(vel, a, b)

    popt, _ = curve_fit(power_bowers, sigma_vc/sigma_max, es/sigma_max)

    u, = popt

    return u


def optimize_bowers_trace(depth_tr, vel_tr, obp_tr, hydro_tr,
                          depth_upper, depth_lower):

    es_data = np.array(obp_tr) - np.array(hydro_tr)
    depth = np.array(depth_tr)

    mask = depth < depth_lower
    mask *= depth > depth_upper

    vel_interval = np.array(vel_tr)[mask]
    es_interval = es_data[mask]

    vel_to_fit = pick_sparse(vel_interval, 3)
    es_to_fit = pick_sparse(es_interval, 3)

    popt, _ = curve_fit(virgin_curve, es_to_fit, vel_to_fit)
    a, b = popt

    return a, b


def optimize_eaton(well, vel_log, obp_log, a, b, pres_log="loading"):
    """
    Optimizer for Eaton model

    Parameters
    ----------
    well : Well
    vel_log : Log or str
        Log object or well log name stored in well
    obp_log : Log or str
        Log object or well log name stored in well
    a, b : float
        coefficients of NCT
    pres_log : Log or str
        Log object storing measured pressure value or Pressure name
        stored in well

    Returns
    -------
    n : float
        optimized eaton exponential
    min_eer : float
        minimum error abtained by optimized n
    rms_err : array
        array of rms error of different n around minmum
    """
    if isinstance(vel_log, (bytes, str)):
        vel_log = well.get_log(vel_log)
    if isinstance(obp_log, (bytes, str)):
        obp_log = well.get_log(obp_log)
    if isinstance(pres_log, (bytes, str)):
        # pres_log = well.get_loading_pressure()
        pres_log = well.get_pressure(pres_log)

    depth = np.array(obp_log.depth)

    hydrostatic = np.array(well.hydrostatic)
    es_normal = np.array(obp_log.data) - hydrostatic
    v_normal = normal(depth, a, b)

    vel = list()
    vel_norm = list()
    es_norm = list()
    obp = list()
    pres = list()
    for dp in pres_log.depth:
        idx = np.searchsorted(depth, dp)
        vel.append(vel_log.data[idx])
        vel_norm.append(v_normal[idx])
        es_norm.append(es_normal[idx])
        obp.append(obp_log.data[idx])
    vel, vel_norm = np.array(vel), np.array(vel_norm)
    vel_ratio = vel / vel_norm

    obp, pres = np.array(obp), np.array(pres_log.data)
    es = obp - pres
    es_norm = np.array(es_norm)
    es_ratio = es / es_norm

    popt, _ = curve_fit(power_eaton, vel_ratio, es_ratio)
    n, = popt

    return n


def optimize_multivaraite(well, obp_log, vel_log, por_log, vsh_log, B,
                          upper, lower):
    if isinstance(vel_log, (bytes, str)):
        vel_log = well.get_log(vel_log)
    if isinstance(por_log, (bytes, str)):
        por_log = well.get_log(por_log)
    if isinstance(vsh_log, (bytes, str)):
        vsh_log = well.get_log(vsh_log)
    if isinstance(obp_log, (bytes, str)):
        obp_log = well.get_log(obp_log)
    # --------------------------------------------
    obp_data = np.array(obp_log.data)
    por_data = np.array(por_log.data)
    vp_data = np.array(vel_log.data)
    vsh_data = np.array(vsh_log.data)

    depth = well.depth
    hydrostatic = well.hydrostatic

    es = obp_data - hydrostatic

    B = well.params['bowers']['B'] if B is None else B  # 0.88
    es_data = es**B

    mask = np.isfinite(vp_data)
    mask *= np.isfinite(por_data)
    mask *= np.isfinite(vsh_data)
    mask *= depth < lower #3500
    mask *= depth > upper #1500

    por_data = por_data[mask]
    vsh_data = vsh_data[mask]
    es_data = es_data[mask]
    vp_data = vp_data[mask]
    depth = depth[mask]

    por_data = por_data.reshape((por_data.shape[0], 1))
    vsh_data = vsh_data.reshape((vsh_data.shape[0], 1))
    es_data = es_data.reshape((es_data.shape[0], 1))
    vp_data = vp_data.reshape((vp_data.shape[0], 1))

    #_____________________REGRESSION______________________
    data = np.hstack((por_data, vsh_data, es_data))
    # global REG
    reg = linear_model.LinearRegression()
    reg.fit(data, vp_data)

    r2 = reg.score(data, vp_data)

    a0 = reg.intercept_[0]
    a1 = -reg.coef_[0][0]
    a2 = -reg.coef_[0][1]
    a3 = reg.coef_[0][2]

    return a0, a1, a2, a3


def optimize_nct(vel_log, fit_start, fit_stop):
    """
    Fit velocity NCT

    Parameters
    ----------
    vel_log : Log
        Velocity log
    fit_start, fit_stop : float
        start and end depth for fitting

    Returns
    -------
    a, b : float
        NCT coefficients
    """
    fit_start = fit_start
    fit_stop = fit_stop

    if fit_start is None or fit_start < vel_log.top:
        fit_start = vel_log.top
    if fit_stop is None or fit_stop > vel_log.bottom:
        fit_stop = vel_log.bottom

    a, b = optimize_nct_trace(
        np.array(vel_log.depth), np.array(vel_log.data), fit_start, fit_stop,
        pick=False)

    return a, b


def optimize_nct_trace(depth, vel, fit_start, fit_stop, pick=True):

    mask = depth > fit_start
    mask *= depth < fit_stop
    mask *= np.isfinite(vel)

    depth_interval = np.array(depth)[mask]
    vel_interval = np.array(vel)[mask]

    if pick is True:
        depth_to_fit = pick_sparse(depth_interval, 3)
        vel_to_fit = pick_sparse(vel_interval, 3)
    else:
        depth_to_fit = depth_interval
        vel_to_fit = vel_interval

    dt = vel_to_fit**(-1)
    log_dt = np.log(dt)

    popt, _ = curve_fit(normal_dt, depth_to_fit, log_dt)
    a, b = popt

    return a, b


def optimize_traugott(den_log, fit_start, fit_stop, kb=0, wd=0):
    """
    Fit density variation against depth with Traugott equation

    Parameters
    ----------
    den_log : Log
        Density log
    fit_start, fit_stop : float
        start and end depth for fitting
    kb : float
        kelly bushing height in meters
    wd : float
        water depth in meters

    Returns
    -------
    a, b : float
        Traugott equation coefficients
    """
    start_idx = den_log.get_depth_idx(fit_start)
    stop_idx = den_log.get_depth_idx(fit_stop) + 1
    depth = np.array(den_log.depth[start_idx: stop_idx + 1])
    den = np.array(den_log.data[start_idx: stop_idx + 1])

    mask = np.isfinite(den)
    den_finite = den[mask]
    depth_finite = depth[mask]

    depth_finite_shift = depth_finite - kb - wd

    popt, _ = curve_fit(traugott, depth_finite_shift, den_finite)
    a, b = popt

    return a, b
